import jax.numpy as jnp
import pytest

from apax.layers.empirical import LatentEwald


@pytest.fixture
def setup_data():
    n_atoms = 4

    R = jnp.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
    dr_vec = None
    Z = jnp.array([1, 1, 1, 1])
    idx = None
    box = jnp.array([[3.0, 0.0, 0.0], [0.0, 3.0, 0.0], [0.0, 0.0, 3.0]])
    properties = {
        "charge": jnp.array([1.0, -1.0, 1.0, -1.0])[:, None]
    }  # Charges on atoms

    return R, dr_vec, Z, idx, box, properties


def test_latent_ewald(setup_data):
    R, dr_vec, Z, idx, box, properties = setup_data

    latent_ewald = LatentEwald(apply_mask=True, sigma=1.0, kgrid=[2, 2, 2])

    # Check for KeyError if "charge" property is missing
    with pytest.raises(KeyError, match="property 'charge' not found"):
        latent_ewald(R, dr_vec, Z, idx, box, {})

    energy = latent_ewald(R, dr_vec, Z, idx, box, properties)

    # Validate the energy output
    assert isinstance(energy, jnp.ndarray), "Output should be a JAX array"
    assert energy.ndim == 0, "Energy should be a scalar value"

    # Test with a larger kgrid
    latent_ewald = LatentEwald(apply_mask=False, sigma=1.5, kgrid=[4, 4, 4])
    energy = latent_ewald(R, dr_vec, Z, idx, box, properties)
    assert isinstance(energy, jnp.ndarray)



import flax.linen as nn
import jax.numpy as jnp

from apax.utils.convert import str_to_dtype


class NTKLinear(nn.Module):
    """Linear layer with activation.
    Corresponds to an NTK layer with "normal" and "zeros" for w and b initialization
    and "use_ntk" set to True.
    """

    units: int
    w_init: str = "normal"
    b_init: str = "zeros"
    use_ntk: bool = True
    dtype: str = "fp32"

    @nn.compact
    def __call__(self, inputs):
        dtype = str_to_dtype(self.dtype)
        inputs = inputs.astype(dtype)

        if self.w_init == "normal":
            w_initializer = nn.initializers.normal(1.0, dtype=dtype)
        elif self.w_init == "lecun":
            w_initializer = nn.initializers.lecun_normal(dtype=dtype)
        else:
            raise ValueError(f"Unknown weight initializer: {self.w_init}.")

        if self.b_init == "normal":
            b_initializer = nn.initializers.normal(1.0, dtype=dtype)
        elif self.b_init == "zeros":
            b_initializer = nn.initializers.constant(0.0, dtype=dtype)
        else:
            raise ValueError(f"Unknown bias initializer: {self.b_init}.")
        w = self.param("w", w_initializer, (inputs.shape[0], self.units), dtype)
        b = self.param("b", b_initializer, [self.units], dtype)

        wx = jnp.dot(inputs, w)

        if self.use_ntk:
            bias_factor = 0.1
            weight_factor = jnp.sqrt(1.0 / inputs.shape[0])
            prediction = weight_factor * wx + bias_factor * b
        else:
            prediction = wx + b

        assert prediction.dtype == dtype
        return prediction



import numpy as np
from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator

from apax.data.statistics import PerElementRegressionShift
from apax.utils.convert import atoms_to_inputs, atoms_to_labels


def test_energy_per_element():
    pos = [0.0, 0.0, 0.0]
    atoms1 = Atoms("H2O", positions=[pos, pos, pos])
    atoms2 = Atoms("CH4", positions=[pos, pos, pos, pos, pos])
    atoms3 = Atoms("N2", positions=[pos, pos])

    dummy_energies = np.array([0, 20.0, 0, 0, 0, 0, 50.0, 100.0, 70.0])

    atoms_list = [atoms1, atoms2, atoms3]
    energies = []
    n_atoms = []
    for atoms in atoms_list:
        n_atoms.append(len(atoms))
        energy = np.sum(dummy_energies[atoms.numbers])
        energies.append(energy)
        atoms.calc = SinglePointCalculator(atoms, energy=energy)

    inputs = atoms_to_inputs(
        atoms_list,
    )
    labels = atoms_to_labels(atoms_list)

    elemental_shift = PerElementRegressionShift.compute(
        inputs, labels, {"energy_regularisation": 0.0}
    )
    regression_energies = []
    for atoms in atoms_list:
        energy = np.sum(elemental_shift[atoms.numbers])
        regression_energies.append(energy)

    assert np.allclose(energies, regression_energies, atol=1e-4)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Describes different physical quantities."""

import functools
import operator
from typing import Callable, Optional, Tuple, TypeVar, Union

import jax.numpy as jnp
from jax import ShapeDtypeStruct, eval_shape, grad, ops, tree_util, vmap
from jax.scipy.special import gammaln

from apax.utils.jax_md_reduced import dataclasses, partition, space, util

partial = functools.partial

# Types


Array = util.Array
f32 = util.f32
f64 = util.f64

DisplacementFn = space.DisplacementFn
MetricFn = space.MetricFn
Box = space.Box

EnergyFn = Callable[..., Array]
ForceFn = Callable[..., Array]

T = TypeVar("T")
InitFn = Callable[..., T]
ApplyFn = Callable[[T], T]
Simulator = Tuple[InitFn, ApplyFn]


# Functions


def force(energy_fn: EnergyFn) -> ForceFn:
    """Computes the force as the negative gradient of an energy."""
    return grad(lambda R, *args, **kwargs: -energy_fn(R, *args, **kwargs))


def clipped_force(energy_fn: EnergyFn, max_force: float) -> ForceFn:
    force_fn = force(energy_fn)

    def wrapped_force_fn(R, *args, **kwargs):
        force = force_fn(R, *args, **kwargs)
        force_norm = jnp.linalg.norm(force, axis=-1, keepdims=True)
        return jnp.where(force_norm > max_force, force / force_norm * max_force, force)

    return wrapped_force_fn


def canonicalize_force(energy_or_force_fn: Union[EnergyFn, ForceFn]) -> ForceFn:
    _force_fn = None

    def force_fn(R, **kwargs):
        nonlocal _force_fn
        if _force_fn is None:
            out_shaped = eval_shape(energy_or_force_fn, R, **kwargs)
            if isinstance(out_shaped, ShapeDtypeStruct) and out_shaped.shape == ():
                _force_fn = force(energy_or_force_fn)
            else:
                # Check that the output has the right shape to be a force.
                is_valid_force = tree_util.tree_reduce(
                    lambda x, y: x and y,
                    tree_util.tree_map(lambda x, y: x.shape == y.shape, out_shaped, R),
                    True,
                )
                if not is_valid_force:
                    raise ValueError(
                        "Provided function should be compatible with "
                        "either an energy or a force. Found a function "
                        f"whose output has shape {out_shaped}."
                    )

                _force_fn = energy_or_force_fn
        return _force_fn(R, **kwargs)

    return force_fn


@functools.singledispatch
def count_dof(position: Array) -> int:
    util.check_custom_simulation_type(position)
    return tree_util.tree_reduce(lambda accum, x: accum + x.size, position, 0)


def volume(dimension: int, box: Box) -> float:
    if jnp.isscalar(box) or not box.ndim:
        return box**dimension
    elif box.ndim == 1:
        return jnp.prod(box)
    elif box.ndim == 2:
        return jnp.linalg.det(box)
    raise ValueError(f"Box must be either: a scalar, a vector, or a matrix. Found {box}.")


def kinetic_energy(
    *unused_args,
    momentum: Array = None,
    velocity: Array = None,
    mass: Array = 1.0,
) -> float:
    """Computes the kinetic energy of a system.

    To avoid ambiguity, either momentum or velocity must be passed explicitly
    as a keyword argument.

    Args:
      momentum: Array specifying the momentum of the system.
      velocity: Array specifying the velocity of the system.
      mass: Array specifying the mass of the constituents.

    Returns:
      The kinetic energy of the system.
    """
    if unused_args:
        raise ValueError(
            "To use the kinetic energy function, you must explicitly "
            "pass either momentum or velocity as a keyword argument."
        )
    if momentum is not None and velocity is not None:
        raise ValueError(
            "To use the kinetic energy function, you must pass either"
            " a momentum or a velocity."
        )

    k = (lambda v, m: v**2 * m) if momentum is None else (lambda p, m: p**2 / m)
    q = velocity if momentum is None else momentum
    util.check_custom_simulation_type(q)

    ke = tree_util.tree_map(lambda m, q: 0.5 * util.high_precision_sum(k(q, m)), mass, q)
    return tree_util.tree_reduce(operator.add, ke, 0.0)


def temperature(
    *unused_args,
    momentum: Array = None,
    velocity: Array = None,
    mass: Array = 1.0,
) -> float:
    """Computes the temperature of a system.

    To avoid ambiguity, either momentum or velocity must be passed explicitly
    as a keyword argument.

    Args:
      momentum: Array specifying the momentum of the system.
      velocity: Array specifying the velocity of the system.
      mass: Array specifying the mass of the constituents.

    Returns:
      The temperature of the system in units of the Boltzmann constant.
    """
    if unused_args:
        raise ValueError(
            "To use the kinetic energy function, you must explicitly "
            "pass either momentum or velocity as a keyword argument."
        )
    if momentum is not None and velocity is not None:
        raise ValueError(
            "To use the kinetic energy function, you must pass either"
            " a momentum or a velocity."
        )

    t = (lambda v, m: v**2 * m) if momentum is None else (lambda p, m: p**2 / m)
    q = velocity if momentum is None else momentum
    util.check_custom_simulation_type(q)

    dof = count_dof(q)

    kT = tree_util.tree_map(lambda m, q: util.high_precision_sum(t(q, m)) / dof, mass, q)
    return tree_util.tree_reduce(operator.add, kT, 0.0)


def pressure(
    energy_fn: EnergyFn, position: Array, box: Box, kinetic_energy: float = 0.0, **kwargs
) -> float:
    """Computes the internal pressure of a system.

    Args:
      energy_fn: A function that computes the energy of the system. This
        function must take as an argument `perturbation` which perturbs the
        box shape. Any energy function constructed using `smap` or in `energy.py`
        with a standard space will satisfy this property.
      position: An array of particle positions.
      box: A box specifying the shape of the simulation volume. Used to infer the
        volume of the unit cell.
      kinetic_energy: A float specifying the kinetic energy of the system.

    Returns:
      A float specifying the pressure of the system.
    """
    dim = position.shape[1]

    def U(eps):
        try:
            return energy_fn(position, box=box, perturbation=(1 + eps), **kwargs)
        except space.UnexpectedBoxException:
            return energy_fn(position, perturbation=(1 + eps), **kwargs)

    dUdV = grad(U)
    vol_0 = volume(dim, box)

    return 1 / (dim * vol_0) * (2 * kinetic_energy - dUdV(0.0))


def stress(
    energy_fn: EnergyFn,
    position: Array,
    box: Box,
    mass: Array = 1.0,
    velocity: Optional[Array] = None,
    **kwargs,
) -> Array:
    """Computes the internal stress of a system.

    Args:
      energy_fn: A function that computes the energy of the system. This
        function must take as an argument `perturbation` which perturbs the
        box shape. Any energy function constructed using `smap` or in `energy.py`
        with a standard space will satisfy this property.
      position: An array of particle positions.
      box: A box specifying the shape of the simulation volume. Used to infer the
        volume of the unit cell.
      mass: The mass of the particles; only used to compute the kinetic
        contribution if `velocity` is not `None`.
      velocity: An array of atomic velocities.

    Returns:
      A float specifying the pressure of the system.
    """
    dim = position.shape[1]

    zero = jnp.zeros((dim, dim), position.dtype)
    I = jnp.eye(dim, dtype=position.dtype)

    def U(eps):
        try:
            return energy_fn(position, box=box, perturbation=(I + eps), **kwargs)
        except space.UnexpectedBoxException:
            return energy_fn(position, perturbation=(I + eps), **kwargs)

    dUdV = grad(U)
    vol_0 = volume(dim, box)

    VxV = 0.0
    if velocity is not None:
        V = velocity
        VxV = util.high_precision_sum(mass * V[:, None, :] * V[:, :, None], axis=0)

    return 1 / vol_0 * (VxV - dUdV(zero))


def cosine_angle_between_two_vectors(dR_12: Array, dR_13: Array) -> Array:
    dr_12 = space.distance(dR_12) + 1e-7
    dr_13 = space.distance(dR_13) + 1e-7
    cos_angle = jnp.dot(dR_12, dR_13) / dr_12 / dr_13
    return jnp.clip(cos_angle, -1.0, 1.0)


def cosine_angles(dR: Array) -> Array:
    """Returns cosine of angles for all atom triplets.

    Args:
      dR: Matrix of displacements; `ndarray(shape=[num_atoms, num_neighbors,
        spatial_dim])`.

    Returns:
      Tensor of cosine of angles;
      `ndarray(shape=[num_atoms, num_neighbors, num_neighbors])`.
    """

    angles_between_all_triplets = vmap(
        vmap(vmap(cosine_angle_between_two_vectors, (0, None)), (None, 0)), 0
    )
    return angles_between_all_triplets(dR, dR)


def is_integer(x: Array) -> bool:
    return x.dtype == jnp.int32 or x.dtype == jnp.int64


def average_pair_correlation_results(gofr, species=None):
    """Calculate species-based averages of pair correlations.

    Average the results of pair_correlation or pair_correlation_neighbor_list,
    appropriately taking species information into account.

    When species=None, gofr is expected to be an array of shape (N,nr), where N is
    the number of species and nr is the number of radii to be considered. The
    average is calculated over all particles, so an array of shape (nr,) is
    returned.

    When species is specified, gofr is expected to be a list of nspecies arrays,
    each of shape (N,nr), where nspecies is the number of unique species types.
    Here, the average is carried out separately for every pair of species, so the
    returned array has shape (nspecies, nspecies, nr).

    Args:
      gofr: array of shape (N,nr) or a list of arrays of shape (N,nr), where nr is
        the number of radii for which :math:`g(r)` is calculated.
      species: Optional. Array of shape (N,) specifying the species of each
        particle.

    Returns:
      An array of shape (nr,) for species=None, otherwise an array of shape
        (nspecies, nspecies, nr), where nspecies is the number of unique species.
    """
    if species is None:
        return jnp.mean(gofr, axis=0)
    species_types = jnp.unique(species)  # note: this returns in sorted order
    return jnp.array(
        [
            [jnp.mean(gofr[si][species == s], axis=0) for s in species_types]
            for si in range(species_types.size)
        ]
    )


def pair_correlation(
    displacement_or_metric: Union[DisplacementFn, MetricFn],
    radii: Array,
    sigma: float,
    species: Array = None,
    eps: float = 1e-7,
    compute_average: bool = False,
):
    """Computes the pair correlation function at a mesh of distances.

    The pair correlation function measures the number of particles at a given
    distance from a central particle. The pair correlation function is defined by

    .. math::
      g(r) = <\sum_{i \\neq j}\delta(r - |r_i - r_j|)>

    We make the approximation,

    .. math::
      \delta(r) \\approx {1 \over \sqrt{2\pi\sigma^2}e^{-r / (2\sigma^2)}}

    Args:
      displacement_or_metric:
        A function that computes the displacement or distance between two points.
      radii: An array of radii at which we would like to compute :math:`g(r)`.
      sigima: A float specifying the width of the approximating Gaussian.
      species: An optional array specifying the species of each particle. If
        species is None then we compute a single :math:`g(r)` for all particles,
        otherwise we compute one :math:`g(r)` for each species.
      eps: A small additive constant used to ensure stability if the radius is
        zero.

    Returns:
      A function `g_fn` that computes the pair correlation function for a
      collection of particles.

    :math:`g(r)` is calculated separately for each particle. For species=None, the
    output of `g_fn` is an array of shape (N, nr), where N is the number of
    particles passed to `g_fn` and nr is the size of radii (the number of points
    at which we calculate :math:`g(r)`. When species is specified, the output is a
    list of nspecies arrays, each of shape (N, nr), where nspecies is the number
    of unique species. If `gofr` is the output of `g_fn`, then gofr[si][i] gives
    the :math:`g(r)` for particle i considering only pair particles of species si.

    Note: when species is specified, the returned list is in the order of the
    sorted unique species indices, not the order in which they appear.

    """
    d = space.canonicalize_displacement_or_metric(displacement_or_metric)
    d = space.map_product(d)

    inv_rad = 1 / (radii + eps)

    def pairwise(dr, dim):
        return jnp.exp(-f32(0.5) * (dr - radii) ** 2 / sigma**2) * inv_rad ** (dim - 1)

    pairwise = vmap(vmap(pairwise, (0, None)), (0, None))

    if species is None:

        def g_fn(R):
            dim = R.shape[-1]
            mask = 1 - jnp.eye(R.shape[0], dtype=R.dtype)
            g_R = jnp.sum(mask[:, :, jnp.newaxis] * pairwise(d(R, R), dim), axis=(1,))
            if compute_average:
                g_R = average_pair_correlation_results(g_R, species)
            return g_R

    else:
        if not (isinstance(species, jnp.ndarray) and is_integer(species)):
            raise TypeError("Malformed species; expecting array of integers.")
        species_types = jnp.unique(species)

        def g_fn(R):
            dim = R.shape[-1]
            g_R = []
            mask = 1 - jnp.eye(R.shape[0], dtype=R.dtype)
            for s in species_types:
                Rs = R[species == s]
                mask_s = mask[:, species == s, jnp.newaxis]
                g_R += [jnp.sum(mask_s * pairwise(d(Rs, R), dim), axis=(1,))]
            if compute_average:
                g_R = average_pair_correlation_results(g_R, species)
            return g_R

    return g_fn


def pair_correlation_neighbor_list(
    displacement_or_metric: Union[DisplacementFn, MetricFn],
    box_size: Box,
    radii: Array,
    sigma: float,
    species: Array = None,
    dr_threshold: float = 0.5,
    eps: float = 1e-7,
    fractional_coordinates: bool = False,
    format: partition.NeighborListFormat = partition.Dense,
    compute_average: bool = False,
):
    """Computes the pair correlation function at a mesh of distances.

    The pair correlation function measures the number of particles at a given
    distance from a central particle. The pair correlation function is defined by

    .. math::
      g(r) = <\sum_{i \\neq j}\delta(r - |r_i - r_j|)>

    We make the approximation,

    .. math::
      \delta(r) \\approx {1 \over \sqrt{2\pi\sigma^2}e^{-r / (2\sigma^2)}}

    This function uses neighbor lists to speed up the calculation.

    Args:
      displacement_or_metric: A function that computes the displacement or
        distance between two points.
      box_size: The size of the box containing the particles.
      radii: An array of radii at which we would like to compute :math:`g(r)`.
      sigima: A float specifying the width of the approximating Gaussian.
      species: An optional array specifying the species of each particle. If
        species is None then we compute a single :math:`g(r)` for all particles,
        otherwise we compute one :math:`g(r)` for each species.
      dr_threshold: A float specifying the halo size of the neighbor list.
      eps: A small additive constant used to ensure stability if the radius is
        zero.
      fractional_coordinates: Bool determining whether positions are stored in
        the unit cube or not.
      format: The format of the neighbor lists. Must be `Dense` or `Sparse`.

    Returns:
      A pair of functions: `neighbor_fn` that constructs a neighbor list (see
      `neighbor_list` in `partition.py` for details). `g_fn` that computes the
      pair correlation function for a collection of particles given their
      position and a neighbor list.

    :math:`g(r)` is calculated separately for each particle. For species=None, the
    output of `g_fn` is an array of shape (N, nr), where N is the number of
    particles passed to `g_fn` and nr is the size of radii (the number of points
    at which we calculate :math:`g(r)`. When species is specified, the output is a
    list of nspecies arrays, each of shape (N, nr), where nspecies is the number
    of unique species. If `gofr` is the output of `g_fn`, then gofr[si][i] gives
    the :math:`g(r)` for particle i considering only pair particles of species si.

    Note: when species is specified, the returned list is in the order of the
    sorted unique species indices, not the order in which they appear.
    """
    metric = space.canonicalize_displacement_or_metric(displacement_or_metric)
    inv_rad = 1 / (radii + eps)

    def pairwise(dr, dim):
        return jnp.exp(-f32(0.5) * (dr - radii) ** 2 / sigma**2) * inv_rad ** (dim - 1)

    neighbor_fn = partition.neighbor_list(
        displacement_or_metric,
        box_size,
        jnp.max(radii) + sigma,
        dr_threshold,
        format=format,
    )

    if species is None:

        def g_fn(R, neighbor):
            N, dim = R.shape
            mask = partition.neighbor_list_mask(neighbor)
            if neighbor.format is partition.Dense:
                R_neigh = R[neighbor.idx]
                d = space.map_neighbor(metric)
                _pairwise = vmap(vmap(pairwise, (0, None)), (0, None))
                g_R = jnp.sum(mask[:, :, None] * _pairwise(d(R, R_neigh), dim), axis=(1,))
                if compute_average:
                    g_R = average_pair_correlation_results(g_R, species)
                return g_R
            elif neighbor.format is partition.Sparse:
                dr = space.map_bond(metric)(R[neighbor.idx[0]], R[neighbor.idx[1]])
                _pairwise = vmap(pairwise, (0, None))
                g_R = ops.segment_sum(
                    mask[:, None] * _pairwise(dr, dim), neighbor.idx[0], N
                )
                if compute_average:
                    g_R = average_pair_correlation_results(g_R, species)
                return g_R
            else:
                raise NotImplementedError(
                    "Pair correlation function does not support "
                    "OrderedSparse neighbor lists."
                )

    else:
        if not (isinstance(species, jnp.ndarray) and is_integer(species)):
            raise TypeError("Malformed species; expecting array of integers.")
        species_types = jnp.unique(species)

        def g_fn(R, neighbor):
            N, dim = R.shape
            g_R = []
            mask = partition.neighbor_list_mask(neighbor)
            if neighbor.format is partition.Dense:
                neighbor_species = species[neighbor.idx]
                R_neigh = R[neighbor.idx]
                d = space.map_neighbor(metric)
                _pairwise = vmap(vmap(pairwise, (0, None)), (0, None))
                for s in species_types:
                    mask_s = mask * (neighbor_species == s)
                    g_R += [
                        jnp.sum(
                            mask_s[:, :, jnp.newaxis] * _pairwise(d(R, R_neigh), dim),
                            axis=(1,),
                        )
                    ]
            elif neighbor.format is partition.Sparse:
                neighbor_species = species[neighbor.idx[1]]
                dr = space.map_bond(metric)(R[neighbor.idx[0]], R[neighbor.idx[1]])
                _pairwise = vmap(pairwise, (0, None))
                for s in species_types:
                    mask_s = mask * (neighbor_species == s)
                    g_R += [
                        ops.segment_sum(
                            mask_s[:, None] * _pairwise(dr, dim), neighbor.idx[0], N
                        )
                    ]
            else:
                raise NotImplementedError(
                    "Pair correlation function does not support "
                    "OrderedSparse neighbor lists."
                )

            if compute_average:
                g_R = average_pair_correlation_results(g_R, species)
            return g_R

    return neighbor_fn, g_fn


def nball_unit_volume(spatial_dimension: int) -> float:
    """Return the volume of a unit sphere in arbitrary dimensions"""
    return jnp.power(jnp.pi, spatial_dimension / 2) / jnp.exp(
        gammaln(spatial_dimension / 2 + 1)
    )


def particle_volume(
    radii: Array, spatial_dimension: int, particle_count: Array = 1, species: Array = None
) -> float:
    """Calculate the volume of a collection of particles

    Args:
      radii: array of shape (n,) giving particle radii, where n can be 1, the
        number of species, or the number of particles depending on the values of
        particle_count and species.
      spatial_dimension: int giving the spatial dimension
      particle_count: number of particles with each radii. Broadcastable to radii.
      species: list of particle species. If provided, this overrides
        particle_count and radii is expected to give per-species radii

    Returns: the sum of the volume of all the particles
    """
    V_unit = nball_unit_volume(spatial_dimension)
    V_particle = V_unit * radii**spatial_dimension

    if species is not None:
        particle_count = jnp.bincount(species)

    return jnp.sum(particle_count * V_particle)


def volume_fraction(
    box: Box,
    radii: Array,
    spatial_dimension: int,
    particle_count: Array = 1,
    species: Array = None,
) -> float:
    """Calculate the volume fraction

    See documentation for particle_volume for explanation of parameters
    """
    Vparticle = particle_volume(radii, spatial_dimension, particle_count, species)
    return Vparticle / volume(spatial_dimension, box)


def box_size_at_volume_fraction(
    volume_fraction: float,
    radii: Array,
    spatial_dimension: int,
    particle_count: Array = 1,
    species: Array = None,
) -> float:
    """Calculate box_size to obtain a desired volume fraction

    See documentation for particle_volume for explanation of parameters
    """
    Vparticle = particle_volume(radii, spatial_dimension, particle_count, species)
    return jnp.power(Vparticle / volume_fraction, 1 / spatial_dimension)


def box_size_at_number_density(
    particle_count: int, number_density: float, spatial_dimension: int
) -> float:
    return jnp.power(particle_count / number_density, 1 / spatial_dimension)


def box_from_parameters(
    a: float, b: float, c: float, alpha: float, beta: float, gamma: float
) -> Box:
    alpha = alpha * jnp.pi / 180
    beta = beta * jnp.pi / 180
    gamma = gamma * jnp.pi / 180
    yy = b * jnp.sin(gamma)
    xy = b * jnp.cos(gamma)
    xz = c * jnp.cos(beta)
    yz = (b * c * jnp.cos(alpha) - xy * xz) / yy
    zz = jnp.sqrt(c**2 - xz**2 - yz**2)
    return jnp.array([[a, xy, xz], [0, yy, yz], [0, 0, zz]])


def bulk_modulus(elastic_tensor: Array) -> float:
    return jnp.einsum("iijj->", elastic_tensor) / elastic_tensor.shape[0] ** 2


@dataclasses.dataclass
class PHopState:
    position_buffer: jnp.ndarray
    phop: jnp.ndarray


InitFn = Callable[[Array], PHopState]
ApplyFn = Callable[[PHopState, Array], PHopState]
PHopCalculator = Tuple[InitFn, ApplyFn]


def phop(displacement: DisplacementFn, window_size: int) -> PHopCalculator:
    """Computes the phop indicator of rearrangements.

    phop is an indicator function that is effective at detecting when particles
    in a quiescent system have experienced a rearrangement. Qualitatively, phop
    measures when the average position of a particle has changed significantly.

    Formally, given a window of size :math:`\Delta t` we two averages before and after
    the current time,

    .. math::
      E_A[f] = E_{t\in[t - \Delta t / 2, t]}[f(t)]
      E_B[f] = E_{t\in[t, t + \Delta t / 2]}[f(t)].

    In terms of these expectations, phop is given by,

    .. math::
      phop = \sqrt{E_A[(R_i(t) - E_B[R_i(t)])^2]E_B[(R_i(t) - E_A[R_i(t)])^2]}

    phop was first introduced in Candelier et al. [#candelier]_

    Args:
      displacement: A function that computes displacements between pairs of
        particles. See `spaces.py` for details.
      window_size: An integer specifying the number of positions that constitute
        the window.

    Returns:
      A pair of functions, `(init_fn, update_fn)` that initialize the state of a
      phop measurement and update the state of a phop measurement to include new
      positions.

    .. rubric:: References
    .. [#candelier] R. Candelier et al.
      "Spatiotemporal Hierarchy of Relaxation Events, Dynamical Heterogeneities,
      and Structural Reorganization in a Supercooled Liquid"
      Physical Review Letters 105, 135702 (2010).
    """

    half_window_size = window_size // 2
    displacement = space.map_bond(displacement)

    def init_fn(position: Array) -> PHopState:
        position_buffer = jnp.tile(position, (window_size, 1, 1))
        assert position_buffer.shape == ((window_size,) + position.shape)
        return PHopState(
            position_buffer, jnp.zeros((position.shape[0],))
        )  # pytype: disable=wrong-arg-count

    def update_fn(state: PHopState, position: Array) -> PHopState:
        # Compute phop.
        a_pos = state.position_buffer[:half_window_size]
        a_mean = jnp.mean(a_pos, axis=0)
        b_pos = state.position_buffer[half_window_size:]
        b_mean = jnp.mean(b_pos, axis=0)

        phop = jnp.sqrt(
            jnp.mean((a_pos - b_mean) ** 2 * (b_pos - a_mean) ** 2, axis=(0, 2))
        )

        # Unwrap position.
        buff = state.position_buffer
        position = displacement(position, buff[-1]) + buff[-1]

        # Add position to the list.
        buff = jnp.concatenate((buff, position[jnp.newaxis, :, :]))[1:]

        return PHopState(buff, phop)  # pytype: disable=wrong-arg-count

    return init_fn, update_fn



from apax.utils.jax_md_reduced import (
    dataclasses,
    partition,
    quantity,
    simulate,
    smap,
    space,
    util,
)

__all__ = ["partition", "space", "quantity", "simulate", "smap", "util", "dataclasses"]



import pathlib

import pytest
import yaml
from typer.testing import CliRunner

from apax.cli.apax_app import app, template_app, validate_app

TEST_PATH = pathlib.Path(__file__).parent.resolve()

runner = CliRunner()


def test_cli_validate(get_tmp_path):
    # This test also checks whether the templates we provide are valid
    pytest.MonkeyPatch().chdir(get_tmp_path)
    result = runner.invoke(app, ["-h"])
    assert result.exit_code == 0
    assert "Commands" in result.stdout

    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0

    result = runner.invoke(template_app, "train")
    assert result.exit_code == 0
    result = runner.invoke(template_app, "train")
    assert result.exit_code == 1

    result = runner.invoke(template_app, ["train", "--full"])
    assert result.exit_code == 0
    result = runner.invoke(template_app, ["train", "--full"])
    assert result.exit_code == 1

    result = runner.invoke(template_app, "md")
    assert result.exit_code == 0
    result = runner.invoke(template_app, "md")
    assert result.exit_code == 1

    # Expect the unedited templates to fail
    result = runner.invoke(validate_app, ["train", "config.yaml"])
    assert result.exit_code == 1

    result = runner.invoke(validate_app, ["train", "config_full.yaml"])
    assert result.exit_code == 1

    result = runner.invoke(validate_app, ["md", "md_config.yaml"])
    assert result.exit_code == 1

    # Load and fix the train config, then try again
    with open("config_full.yaml", "r") as stream:
        model_config_dict = yaml.safe_load(stream)

    model_config_dict["n_epochs"] = 100
    model_config_dict["data"]["data_path"] = "dataset.extxyz"

    with open("config_full_fixed.yaml", "w") as conf:
        yaml.dump(model_config_dict, conf, default_flow_style=False)

    result = runner.invoke(validate_app, ["train", "config_full_fixed.yaml"])
    assert result.exit_code == 0
    assert "Success!" in result.stdout

    result = runner.invoke(app, ["visualize", "config_full_fixed.yaml"])
    assert result.exit_code == 0
    assert "Total Parameters" in result.stdout

    # same for md
    with open("md_config.yaml", "r") as stream:
        md_config_dict = yaml.safe_load(stream)

    md_config_dict["initial_structure"] = "initial_structure.extxyz"
    md_config_dict["duration"] = 10
    md_config_dict["ensemble"]["temperature_schedule"]["T0"] = 10

    with open("md_config_fixed.yaml", "w") as conf:
        yaml.dump(md_config_dict, conf, default_flow_style=False)

    result = runner.invoke(validate_app, ["md", "md_config_fixed.yaml"])
    assert result.exit_code == 0
    assert "Success!" in result.stdout



import logging
import typing
from pathlib import Path

import ase.io
import numpy as np
import zntrack.utils
from matplotlib import pyplot as plt

try:
    from sklearn.decomposition import PCA
except ImportError:
    PCA = None

from apax.bal import kernel_selection
from apax.nodes.model import ApaxBase

log = logging.getLogger(__name__)


def plot_with_uncertainty(value, ylabel: str, xlabel: str, x=None, **kwargs) -> tuple:
    """Parameters
    ----------
    value: data of shape (n, m) where n is the number of ensembles.
    x: optional x values of shape (m,)

    Returns
    -------

    """
    if isinstance(value, dict):
        data = value
    else:
        data = {
            "mean": np.mean(value, axis=0),
            "std": np.std(value, axis=0),
            "max": np.max(value, axis=0),
            "min": np.min(value, axis=0),
        }

    fig, ax = plt.subplots(**kwargs)
    if x is None:
        x = np.arange(len(data["mean"]))
    ax.fill_between(
        x,
        data["mean"] + data["std"],
        data["mean"] - data["std"],
        facecolor="lightblue",
    )
    if "max" in data:
        ax.plot(x, data["max"], linestyle="--", color="darkcyan")
    if "min" in data:
        ax.plot(x, data["min"], linestyle="--", color="darkcyan")
    ax.plot(x, data["mean"], color="black")
    ax.set_ylabel(ylabel)
    ax.set_xlabel(xlabel)

    return fig, ax, data


class BatchKernelSelection(zntrack.Node):
    """Interface to the batch active learning methods implemented in apax.
    Check the apax documentation for a list and explanation of implemented properties.

    Attributes
    ----------
    models: Union[Apax, List[Apax]]
        One or more Apax models to construct a feature map from.
    base_feature_map: dict
        Name and parameters for the feature map transformation.
    selection_method: str
        Name of the selection method to be used. Choose from:
        ["max_dist", ]
    n_configurations: int
        Number of samples to be selected.
    min_distance_threshold: float
        Minimal allowed distance between selected points before selection stops.
        Selection stops either if n samples were selected or if the distance is smaller
        than specified by this criterion.
        Default is 0 so this criterion will not apply.
    rank_all: bool
        If true, ranks all pool samples. Great for visualization but might slow things
        down if the pool is really large. If false, only the most informative
        n_configurations are ranked.
    processing_batch_size: int
        Number of samples to be processed in parallel.
        Does not affect the result, just the speed of computing features.
    """

    data: list[ase.Atoms] = zntrack.deps()
    train_data: list[ase.Atoms] = zntrack.deps()

    selected_ids: list[int] = zntrack.outs(independent=True)

    models: typing.Union[ApaxBase, typing.List[ApaxBase]] = zntrack.deps()
    base_feature_map: dict = zntrack.params(
        default_factory=lambda: {"name": "ll_grad", "layer_name": "dense_2"}
    )
    selection_method: str = zntrack.params("max_dist")
    n_configurations: int = zntrack.params()
    min_distance_threshold: int = zntrack.params(0.0)
    processing_batch_size: int = zntrack.params(64)
    rank_all: bool = zntrack.params(True)

    img_selection: Path = zntrack.outs_path(zntrack.nwd / "selection.png")
    img_distances: Path = zntrack.outs_path(zntrack.nwd / "distances.png")
    img_features: Path = zntrack.outs_path(zntrack.nwd / "features.png")

    def get_data(self) -> list[ase.Atoms]:
        """Get the atoms data to process."""
        if self.data is not None:
            return self.data
        else:
            raise ValueError("No data given.")

    def run(self):
        """ZnTrack Node Run method."""

        log.debug(f"Selecting from {len(self.data)} configurations.")
        self.selected_ids = self.select_atoms(self.data)

    @property
    def frames(self) -> list[ase.Atoms]:
        """Get a list of the selected atoms objects."""
        return [atoms for i, atoms in enumerate(self.data) if i in self.selected_ids]

    @property
    def excluded_frames(self) -> list[ase.Atoms]:
        """Get a list of the atoms objects that were not selected."""
        return [atoms for i, atoms in enumerate(self.data) if i not in self.selected_ids]

    def select_atoms(self, atoms_lst: typing.List[ase.Atoms]) -> typing.List[int]:
        if isinstance(self.models, list):
            param_files = [m.parameter["data"]["directory"] for m in self.models]
        else:
            param_files = self.models.parameter["data"]["directory"]

        if self.rank_all:
            selection_batch_size = None
        else:
            selection_batch_size = self.n_configurations

        ranking, distances, features_train, features_pool = kernel_selection(
            param_files,
            self.train_data,
            atoms_lst,
            self.base_feature_map,
            self.selection_method,
            selection_batch_size=selection_batch_size,
            processing_batch_size=self.processing_batch_size,
        )

        features_pool = features_pool[ranking]
        distance_mask = distances > self.min_distance_threshold

        mask = distance_mask
        if self.rank_all:
            # if we rank all, we need to slice the indices to only
            # select the first n
            numbers = np.arange(len(distances))
            number_mask = numbers < self.n_configurations
            mask = mask & number_mask

        ranking = ranking[mask]
        features_selection = features_pool[mask]
        features_remaining = features_pool[~mask]

        false_indices = np.where(~mask)[0]
        if len(false_indices) > 0:
            last_selected = false_indices[0] - 1
        else:
            last_selected = -1

        self._get_distances_plot(distances, last_selected)
        self._get_pca_plot(features_train, features_selection, features_remaining)
        self._get_selection_plot(atoms_lst, ranking)
        return [int(x) for x in ranking]

    def _get_selection_plot(
        self, atoms_lst: typing.List[ase.Atoms], indices: typing.List[int]
    ):
        energies = np.array([atoms.calc.results["energy"] for atoms in atoms_lst])

        if "energy_uncertainty" in atoms_lst[0].calc.results.keys():
            uncertainty = np.array(
                [atoms.calc.results["energy_uncertainty"] for atoms in atoms_lst]
            )
            fig, ax, _ = plot_with_uncertainty(
                {"mean": energies, "std": uncertainty},
                ylabel="energy",
                xlabel="configuration",
            )
        else:
            fig, ax = plt.subplots()
            ax.plot(energies, label="energy")
            ax.set_ylabel("energy")
            ax.set_xlabel("configuration")

        ax.plot(indices, energies[indices], "x", color="red")

        fig.savefig(self.img_selection, bbox_inches="tight", dpi=240)

    def _get_distances_plot(self, distances: np.ndarray, last_selected: int):
        fig, ax = plt.subplots()
        ax.semilogy(distances, label="sq. distances")

        ax.axvline(last_selected, ls="--", color="gray", label="last selected")

        ax.set_ylabel("Squared distance")
        ax.set_xlabel("Configuration")
        ax.legend()
        fig.savefig(self.img_distances, bbox_inches="tight", dpi=240)

    def _get_pca_plot(
        self, g_train: np.ndarray, g_selection: np.ndarray, g_remaining: np.ndarray
    ):
        if PCA:
            all_features = [g_train, g_selection]
            if len(g_remaining) > 0:
                all_features.append(g_remaining)
            g_full = np.concatenate(all_features, axis=0)
            pca = PCA(n_components=2)
            pca.fit(g_full)

            g_train_2d = pca.transform(g_train)
            g_selection_2d = pca.transform(g_selection)
            if len(g_remaining) > 0:
                g_pool_2d = pca.transform(g_remaining)
        else:
            F = g_train.shape[1]
            W = np.random.randn(F, 2) / np.sqrt(F)
            g_train_2d = g_train @ W
            g_selection_2d = g_selection @ W
            if len(g_remaining) > 0:
                g_pool_2d = g_remaining @ W

        fig, ax = plt.subplots()

        if len(g_remaining) > 0:
            ax.scatter(
                *g_pool_2d.T,
                color="gray",
                marker="o",
                s=5,
                alpha=0.6,
                label="Remaining data",
            )

        ax.scatter(
            *g_train_2d.T, color="C0", marker="^", s=5, alpha=0.6, label="Train data"
        )
        ax.scatter(
            *g_selection_2d.T,
            color="red",
            marker="*",
            s=5,
            alpha=0.6,
            label="Selected data",
        )

        ax.set_xlabel("dim 1")
        ax.set_ylabel("dim 2")
        ax.legend()

        fig.savefig(self.img_features, bbox_inches="tight", dpi=240)



import jax
import jax.numpy as jnp

from apax.layers.readout import AtomisticReadout


def test_atomistic_readout():
    key = jax.random.PRNGKey(0)

    x = jnp.linspace(-1, 1, num=10)

    linear = AtomisticReadout([32, 32])
    params = linear.init(key, x)
    result = linear.apply(params, x)

    assert result.shape == (1,)



import jax
import jax.numpy as jnp

from apax.bal.feature_maps import FeatureMap


def ensemble_features(feature_fn: FeatureMap) -> FeatureMap:
    """
    Feature map transformation which averages the kernels of a model ensemble.
    """
    ensemble_feature_fn = jax.vmap(feature_fn, (0, None), 0)

    def averaged_feature_fn(params, x):
        g = ensemble_feature_fn(params, x)

        if len(g.shape) != 2:
            # models, features
            raise ValueError(
                "Dimension mismatch for input features. Expected shape (models,"
                f" features), got {g.shape}"
            )

        n_models = g.shape[0]
        # sqrt since the kernel is K = g^T g
        feature_scale_factor = jnp.sqrt(1 / n_models)
        g_ens = feature_scale_factor * jnp.sum(g, axis=0)  # shape: n_features
        return g_ens

    return averaged_feature_fn


def batch_features(feature_fn: FeatureMap) -> FeatureMap:
    """
    Vectorizes a feature map over structures.
    Should be the last transformation applied to a feature map.
    """
    batched_feature_fn = jax.vmap(feature_fn, (None, 0), 0)
    return batched_feature_fn



import dataclasses

import jax
import jax.numpy as jnp


def make_biased_energy_force_fn(bias_fn):
    def biased_energy_force_fn(positions, Z, idx, box, offsets):
        bias_and_grad_fn = jax.value_and_grad(bias_fn, has_aux=True)

        (E_bias, results), neg_F_bias = bias_and_grad_fn(positions, Z, idx, box, offsets)

        if "energy_unbiased" not in results.keys():
            results["energy_unbiased"] = results["energy"]
            results["forces_unbiased"] = results["forces"]

        F_bias = -neg_F_bias
        results["energy"] = results["energy"] + E_bias
        results["forces"] = results["forces"] + F_bias

        return results

    return biased_energy_force_fn


@dataclasses.dataclass
class UncertaintyDrivenDynamics:
    """
    UDD requires an uncertainty aware model.
    It drives the dynamics towards higher uncertainty regions
    up to some maximum bias energy.
    https://doi.org/10.1038/s43588-023-00406-5


    Parameters
    ----------
    height : float
        Maximum bias potential that can be applied
    width : float
        Width of the Gaussian bias.

    """

    height: float
    width: float

    def apply(self, model):
        def udd_energy(positions, Z, idx, box, offsets):
            n_atoms = positions.shape[0]
            results = model(positions, Z, idx, box, offsets)
            n_models = results["energy_ensemble"].shape[0]

            sigma2 = results["energy_uncertainty"] ** 2

            gauss = jnp.exp(-sigma2 / (n_models * n_atoms * self.width**2))
            E_udd = self.height * (gauss - 1)

            return E_udd, results

        udd_energy_force = make_biased_energy_force_fn(udd_energy)

        return udd_energy_force


@dataclasses.dataclass
class GaussianAcceleratedMolecularDynamics:
    """
    Applies a boost potential to the system that pulls it towards a target energy.
    https://pubs.acs.org/doi/10.1021/acs.jctc.5b00436

    Parameters
    ----------
    energy_target : float
        Target potential energy below which to apply the boost potential.
    spring_constant : float
        Spring constant of the boost potential.
    """

    energy_target: float
    spring_constant: float

    def apply(self, model):
        def gamd_energy(positions, Z, idx, box, offsets):
            results = model(positions, Z, idx, box, offsets)

            energy = jnp.clip(results["energy"], a_max=self.energy_target)

            E_gamd = 0.5 * self.spring_constant * (energy - self.energy_target) ** 2

            return E_gamd, results

        gamd_energy_force = make_biased_energy_force_fn(gamd_energy)

        return gamd_energy_force


@dataclasses.dataclass
class GlobalCalibration:
    """
    Applies a global calibration to energy and force uncertainties.
    Energy ensemble predictions are rescaled according to EQ 7 in
    https://doi.org/10.1063/5.0036522

    Parameters
    ----------
    energy_factor : float
        Global calibration factor by which to scale the energy uncertainty.
    forces_factor : float
        Global calibration factor by which to scale the force uncertainties.
    """

    energy_factor: float
    forces_factor: float

    def apply(self, model):
        def calibrated_model(positions, Z, idx, box, offsets):
            results = model(positions, Z, idx, box, offsets)

            results["energy_uncertainty"] = (
                results["energy_uncertainty"] * self.energy_factor
            )

            Emean = results["energy"]
            Ei = results["energy_ensemble"]
            results["energy_ensemble"] = Emean + self.energy_factor * (Ei - Emean)

            if "forces_uncertainty" in results.keys():
                results["forces_uncertainty"] = (
                    results["forces_uncertainty"] * self.forces_factor
                )

            return results

        return calibrated_model


available_transformations = {
    "udd": UncertaintyDrivenDynamics,
    "gamd": GaussianAcceleratedMolecularDynamics,
    "global_cal": GlobalCalibration,
}



import os
import pathlib
import subprocess
import sys

try:
    import ipsuite as ips
except ImportError:
    pass

import numpy as np
import pytest
import yaml
import zntrack

import apax.nodes
from apax.nodes.model import Apax, ApaxEnsemble
from apax.nodes.utils import AddData

CONFIG_PATH = pathlib.Path(__file__).parent / "example.yaml"
TEST_PATH = pathlib.Path(__file__).parent.resolve()


def save_config_with_seed(path: str, seed: int = 1) -> None:
    with open(CONFIG_PATH, "r") as f:
        config = yaml.safe_load(f)

    config["seed"] = seed

    with open(path, "w") as f:
        yaml.dump(config, f)


def test_n_train_model(tmp_path, get_md22_stachyose):
    os.chdir(tmp_path)
    subprocess.run(["git", "init"], check=True)
    subprocess.run(["dvc", "init"], check=True)
    save_config_with_seed(tmp_path / "example.yaml")
    proj = zntrack.Project()
    with proj:
        data = AddData(file=get_md22_stachyose, stop=100)
        model = Apax(data=data.frames, validation_data=data.frames, config="example.yaml")

    proj.repro()

    model = model.from_rev()
    data = data.from_rev()

    atoms = data.frames[0]
    atoms.calc = model.get_calculator()

    assert atoms.get_potential_energy() < 0


@pytest.mark.skipif("ipsuite" not in sys.modules, reason="requires new ipsuite release")
def test_n_train_2_model(tmp_path, get_md22_stachyose):
    os.chdir(tmp_path)
    subprocess.run(["git", "init"], check=True)
    subprocess.run(["dvc", "init"], check=True)

    save_config_with_seed(tmp_path / "example.yaml")
    save_config_with_seed(tmp_path / "example2.yaml", seed=2)

    proj = zntrack.Project(automatic_node_names=True)
    thermostat = ips.LangevinThermostat(time_step=1.0, temperature=100.0, friction=0.01)
    with proj:
        data = AddData(file=get_md22_stachyose, stop=100)
        model1 = Apax(
            data=data.frames, validation_data=data.frames, config="example.yaml"
        )
        model2 = Apax(
            data=data.frames, validation_data=data.frames, config="example2.yaml"
        )
        ensemble = ApaxEnsemble(models=[model1, model2])
        md = ips.ASEMD(
            data=data.frames,
            model=ensemble,
            thermostat=thermostat,
            steps=20,
            sampling_rate=1,
        )

        uncertainty_selection = ips.ThresholdSelection(
            data=md.frames, n_configurations=1, threshold=0.0001
        )

        selection_batch_size = 3
        kernel_selection = apax.nodes.BatchKernelSelection(
            data=md.frames,
            train_data=data.frames,
            models=[model1, model2],
            n_configurations=selection_batch_size,
            processing_batch_size=4,
        )

        prediction = ips.Prediction(data=kernel_selection.frames, model=ensemble)
        analysis = ips.PredictionMetrics(x=kernel_selection.frames, y=prediction.frames)

    proj.repro()

    model = ensemble.from_rev()
    data = data.from_rev()

    atoms = data.frames[0]
    atoms.calc = model.get_calculator()

    assert atoms.get_potential_energy() < 0

    uncertainties = [x.calc.results["energy_uncertainty"] for x in md.frames]
    assert [md.frames[np.argmax(uncertainties)]] == uncertainty_selection.frames

    assert len(kernel_selection.frames) == selection_batch_size



from apax.layers.descriptor.equiv_mp import EquivMPRepresentation
from apax.layers.descriptor.gaussian_moment_descriptor import GaussianMomentDescriptor

try:
    from apax.layers.descriptor.so3krates import So3kratesRepresentation
except ImportError:
    So3kratesRepresentation = None

__all__ = ["GaussianMomentDescriptor", "EquivMPRepresentation", "So3kratesRepresentation"]



import dataclasses
from typing import List

import jax
import jax.numpy as jnp
import jax.scipy as jsc
import numpy as np

from apax.utils.math import normed_dotp


def weighted_squared_error(
    label: jnp.array,
    prediction: jnp.array,
    name,
    parameters: dict = {},
) -> jnp.array:
    """
    Squared error function that allows weighting of
    individual contributions by the number of atoms in the system.
    """
    label, prediction = label[name], prediction[name]
    return (label - prediction) ** 2


def weighted_huber_loss(
    label: jnp.array,
    prediction: jnp.array,
    name,
    parameters: dict = {},
) -> jnp.array:
    """
    Huber loss function that allows weighting of
    individual contributions by the number of atoms in the system.
    """
    label, prediction = label[name], prediction[name]
    if "delta" not in parameters.keys():
        raise KeyError("Huber loss function requires 'delta' parameter")
    delta = parameters["delta"]
    diff = jnp.abs(label - prediction)
    loss = jnp.where(diff > delta, delta * (diff - 0.5 * delta), 0.5 * diff**2)
    return loss


def crps_loss(
    label: jax.Array,
    prediction: jax.Array,
    name,
    parameters: dict = {},
) -> jax.Array:
    """Computes the CRPS of a gaussian distribution given
    means, targets and standard deviations (uncertainty estimate)
    """
    label = label[name]
    means = prediction[name]
    sigmas = prediction[name + "_uncertainty"]

    sigmas = jnp.clip(sigmas, a_min=1e-6)

    norm_x = (label - means) / sigmas
    cdf = 0.5 * (1 + jsc.special.erf(norm_x / jnp.sqrt(2)))

    normalization = 1 / (jnp.sqrt(2.0 * np.pi))

    pdf = normalization * jnp.exp(-(norm_x**2) / 2.0)

    crps = sigmas * (norm_x * (2 * cdf - 1) + 2 * pdf - 1 / jnp.sqrt(np.pi))

    return crps


def nll_loss(
    label: jax.Array,
    prediction: jax.Array,
    name,
    parameters: dict = {},
) -> jax.Array:
    """Computes the gaussian NLL loss given
    means, targets and standard deviations (uncertainty estimate)
    """
    label = label[name]
    means = prediction[name]
    sigmas = prediction[name + "_uncertainty"]

    eps = 1e-6
    sigmas = jnp.clip(sigmas, a_min=eps)
    variances = jnp.pow(sigmas, 2)

    x1 = jnp.log(variances)
    x2 = ((means - label) ** 2) / variances
    nll = 0.5 * (x1 + x2)

    return nll


def force_angle_loss(
    label: jnp.array,
    prediction: jnp.array,
    name,
    parameters: dict = {},
) -> jnp.array:
    """
    Consine similarity loss function. Contributions are summed in `Loss`.
    """
    label, prediction = label[name], prediction[name]
    dotp = normed_dotp(label, prediction)
    return 1.0 - dotp


def force_angle_div_force_label(
    label: jnp.array,
    prediction: jnp.array,
    name,
    parameters: dict = {},
):
    """
    Consine similarity loss function weighted by the norm of the force labels.
    Contributions are summed in `Loss`.
    """
    label, prediction = label[name], prediction[name]
    dotp = normed_dotp(label, prediction)
    F_0_norm = jnp.linalg.norm(label, ord=2, axis=2, keepdims=False)
    loss = jnp.where(F_0_norm > 1e-6, (1.0 - dotp) / F_0_norm, jnp.zeros_like(dotp))
    return loss


def force_angle_exponential_weight(
    label: jnp.array,
    prediction: jnp.array,
    name,
    parameters: dict = {},
) -> jnp.array:
    """
    Consine similarity loss function exponentially scaled by the norm of the force labels.
    Contributions are summed in `Loss`.
    """
    label, prediction = label[name], prediction[name]
    dotp = normed_dotp(label, prediction)
    F_0_norm = jnp.linalg.norm(label, ord=2, axis=2, keepdims=False)
    return (1.0 - dotp) * jnp.exp(-F_0_norm)


def stress_tril(label, prediction, name, parameters: dict = {}):
    label, prediction = label[name], prediction[name]
    idxs = jnp.tril_indices(3)
    label_tril = label[:, idxs[0], idxs[1]]
    prediction_tril = prediction[:, idxs[0], idxs[1]]
    return (label_tril - prediction_tril) ** 2


loss_functions = {
    "mse": weighted_squared_error,
    "huber": weighted_huber_loss,
    "cosine_sim": force_angle_loss,
    "cosine_sim_div_magnitude": force_angle_div_force_label,
    "cosine_sim_exp_magnitude": force_angle_exponential_weight,
    "tril": stress_tril,
    "crps": crps_loss,
    "nll": nll_loss,
}


@dataclasses.dataclass
class Loss:
    """
    Represents a single weighted loss function that is constructed from a `name`
    and a type of comparison metric.
    """

    name: str
    loss_type: str
    weight: float = 1.0
    atoms_exponent: float = 1.0
    parameters: dict = dataclasses.field(default_factory=lambda: {})

    def __post_init__(self):
        if self.loss_type not in loss_functions.keys():
            raise NotImplementedError(
                f"the loss function '{self.loss_type}' is not known."
            )

        self.loss_fn = loss_functions[self.loss_type]

    def __call__(self, inputs: dict, prediction: dict, label: dict) -> float:
        # TODO we may want to insert an additional `mask` argument for this method

        divisor = inputs["n_atoms"] ** self.atoms_exponent
        batch_losses = self.loss_fn(label, prediction, self.name, self.parameters)

        axes_to_add = len(batch_losses.shape) - 1
        for _ in range(axes_to_add):
            divisor = divisor[..., None]

        arg = batch_losses / divisor
        loss = self.weight * jnp.sum(jnp.mean(arg, axis=0))
        return loss


@dataclasses.dataclass
class LossCollection:
    loss_list: List[Loss]

    def __call__(self, inputs: dict, predictions: dict, labels: dict) -> float:
        total_loss = 0.0
        for single_loss_fn in self.loss_list:
            loss = single_loss_fn(inputs, predictions, labels)
            total_loss = total_loss + loss

        return total_loss



import numpy as np

from apax.train.eval import get_test_idxs


def test_get_test_idxs():
    atoms_list = list(range(10))
    used_idxs = np.array([1, 3, 5, 7, 9])
    test_idxs = get_test_idxs(atoms_list=atoms_list, used_idxs=used_idxs, n_test=2)
    assert len(test_idxs) == 2
    assert not any(item in used_idxs for item in test_idxs)

    test_idxs = get_test_idxs(atoms_list=atoms_list, used_idxs=used_idxs, n_test=-1)
    assert len(test_idxs) == 5
    assert not any(item in used_idxs for item in test_idxs)

    test_idxs = np.append(test_idxs, 7)
    assert any(item in used_idxs for item in test_idxs)



from typing import List, Literal, Optional, Union

from pydantic import (
    BaseModel,
    Field,
    NonNegativeFloat,
    PositiveFloat,
    PositiveInt,
)


class GaussianBasisConfig(BaseModel, extra="forbid"):
    """
    Gaussian primitive basis functions.

    Parameters
    ----------
    n_basis : PositiveInt, default = 7
        Number of uncontracted basis functions.
    r_min : NonNegativeFloat, default = 0.5
        Position of the first uncontracted basis function's mean.
    r_max : PositiveFloat, default = 6.0
        Cutoff radius of the descriptor.
    """

    name: Literal["gaussian"] = "gaussian"
    n_basis: PositiveInt = 7
    r_min: NonNegativeFloat = 0.5
    r_max: PositiveFloat = 6.0


class BesselBasisConfig(BaseModel, extra="forbid"):
    """
    Gaussian primitive basis functions.

    Parameters
    ----------
    n_basis : PositiveInt, default = 16
        Number of uncontracted basis functions.
    r_max : PositiveFloat, default = 5.0
        Cutoff radius of the descriptor.
    """

    name: Literal["bessel"] = "bessel"
    n_basis: PositiveInt = 16
    r_max: PositiveFloat = 5.0


BasisConfig = Union[GaussianBasisConfig, BesselBasisConfig]


class FullEnsembleConfig(BaseModel, extra="forbid"):
    """
    Configuration for full model ensembles.
    Usage can improve accuracy and stability at the cost of slower inference.
    Uncertainties will generally not be calibrated.

    Parameters
    ----------
    n_members : int
        Number of ensemble members.
    """

    kind: Literal["full"] = "full"
    n_members: int


class ShallowEnsembleConfig(BaseModel, extra="forbid"):
    """
    Configuration for shallow (last layer) ensembles.
    Allows use of probabilistic loss functions.
    The predicted uncertainties should be well calibrated.
    See 10.1088/2632-2153/ad594a for details.

    Parameters
    ----------
    n_members : int
        Number of ensemble members.
    force_variance : bool, default = True
        Whether or not to compute force uncertainties.
        Required for probabilistic force loss and calibration of force uncertainties.
        Can lead to better force metrics but but enabling it introduces some non-negligible cost.
    chunk_size : Optional[int], default = None
        If set to an integer, the jacobian of ensemble energies wrt. to positions will be computed
        in chunks of that size. This sacrifices some performance for the possibility to use relatively
        large ensemble sizes.

    Hint
    ----------
    Loss type hase to be changed to a probabilistic loss like 'nll' or 'crps'

    """

    kind: Literal["shallow"] = "shallow"
    n_members: int
    force_variance: bool = True
    chunk_size: Optional[int] = None


EnsembleConfig = Union[FullEnsembleConfig, ShallowEnsembleConfig]


class Correction(BaseModel, extra="forbid"):
    name: str


class ZBLRepulsion(Correction, extra="forbid"):
    name: Literal["zbl"]
    r_max: NonNegativeFloat = 1.5


class ExponentialRepulsion(Correction, extra="forbid"):
    name: Literal["exponential"]
    r_max: NonNegativeFloat = 1.5


class LatentEwald(Correction, extra="forbid"):
    name: Literal["latent_ewald"]
    kgrid: list
    sigma: float = 1.0


EmpiricalCorrection = Union[ZBLRepulsion, ExponentialRepulsion, LatentEwald]


class PropertyHead(BaseModel, extra="forbid"):
    """ """

    name: str
    aggregation: str = "none"
    mode: str = "l0"

    nn: List[PositiveInt] = [128, 128]
    n_shallow_members: int = 0
    w_init: Literal["normal", "lecun"] = "lecun"
    b_init: Literal["normal", "zeros"] = "zeros"
    use_ntk: bool = False
    dtype: Literal["fp32", "fp64"] = "fp32"


class BaseModelConfig(BaseModel, extra="forbid"):
    """
    Configuration for the model.

    Parameters
    ----------
    basis : BasisConfig, default = GaussianBasisConfig()
        Configuration for primitive basis functions.
    nn : List[PositiveInt], default = [256, 256]
        Number of hidden layers and units in those layers.
    w_init : Literal["normal", "lecun"], default = "lecun"
        Initialization scheme for the neural network weights.
    b_init : Literal["normal", "zeros"], default = "zeros"
        Initialization scheme for the neural network biases.
    use_ntk : bool, default = False
        Whether or not to use NTK parametrization.
    ensemble : Optional[EnsembleConfig], default = None
        What kind of model ensemble to use (optional).
    use_zbl : bool, default = False
        Whether to include the ZBL correction.
    calc_stress : bool, default = False
        Whether to calculate stress during model evaluation.
    descriptor_dtype : Literal["fp32", "fp64"], default = "fp32"
        Data type for descriptor calculations.
    readout_dtype : Literal["fp32", "fp64"], default = "fp32"
        Data type for readout calculations.
    scale_shift_dtype : Literal["fp32", "fp64"], default = "fp64"
        Data type for scale and shift parameters.
    """

    basis: BasisConfig = Field(BesselBasisConfig(name="bessel"), discriminator="name")

    nn: List[PositiveInt] = [256, 256]
    w_init: Literal["normal", "lecun"] = "lecun"
    b_init: Literal["normal", "zeros"] = "zeros"
    use_ntk: bool = False

    ensemble: Optional[EnsembleConfig] = None

    property_heads: list[PropertyHead] = []

    # corrections
    empirical_corrections: list[EmpiricalCorrection] = []

    calc_stress: bool = False

    descriptor_dtype: Literal["fp32", "fp64"] = "fp32"
    readout_dtype: Literal["fp32", "fp64"] = "fp32"
    scale_shift_dtype: Literal["fp32", "fp64"] = "fp64"


class GMNNConfig(BaseModelConfig, extra="forbid"):
    """
    Configuration for the model.

    Parameters
    ----------
    n_radial : PositiveInt, default = 5
        Number of contracted basis functions.
    n_contr : int, default = 8
        How many gaussian moment contractions to use.
    emb_init : Optional[str], default = "uniform"
        Initialization scheme for embedding layer weights.
    """

    name: Literal["gmnn"] = "gmnn"

    n_radial: PositiveInt = 5
    n_contr: int = 8
    emb_init: Optional[str] = "uniform"

    def get_builder(self):
        from apax.nn.builder import GMNNBuilder

        return GMNNBuilder


class EquivMPConfig(BaseModelConfig, extra="forbid"):
    """
    Configuration for the model.

    Parameters
    ----------
    features: PositiveInt = 32
        Feature dimension of the linear layers
    max_degree: PositiveInt = 2
        Maximal rotation order for features and tensorproducts
    num_iterations: PositiveInt = 1
        Number of message passing steps.
    """

    name: Literal["equiv-mp"] = "equiv-mp"

    features: PositiveInt = 32
    max_degree: PositiveInt = 2
    num_iterations: PositiveInt = 1

    def get_builder(self):
        from apax.nn.builder import EquivMPBuilder

        return EquivMPBuilder


class So3kratesConfig(BaseModelConfig, extra="forbid"):
    """
    Configuration for the model.

    Parameters
    ----------
    num_layers: PositiveInt = 1
        Number of message passing layers
    max_degree: PositiveInt = 3
        Maximum rotation order
    num_features: PositiveInt = 128
        Feature dimension
    num_heads: PositiveInt = 4
        Number of attention heads
    use_layer_norm_1: bool = False
        Layer norm in transformer block
    use_layer_norm_2: bool = False
        Layer norm in transformer block
    use_layer_norm_final: bool = False
        Layer norm before readout
    activation: str = "silu"
        Activation function
    cutoff_fn: str = "cosine_cutoff"
        Smooth cutoff function
    transform_input_features: bool = False
        Whether or not to apply a dense layer to transformer input features

    """

    name: Literal["so3krates"] = "so3krates"

    num_layers: PositiveInt = 1
    max_degree: PositiveInt = 3
    num_features: PositiveInt = 128
    num_heads: PositiveInt = 4
    use_layer_norm_1: bool = False
    use_layer_norm_2: bool = False
    use_layer_norm_final: bool = False
    activation: str = "silu"
    cutoff_fn: str = "cosine_cutoff"
    transform_input_features: bool = False

    def get_builder(self):
        from apax.nn.builder import So3kratesBuilder

        return So3kratesBuilder


ModelConfig = Union[GMNNConfig, EquivMPConfig, So3kratesConfig]



import dataclasses
import logging

import numpy as np

log = logging.getLogger(__name__)


@dataclasses.dataclass
class DatasetStats:
    elemental_shift: np.array = None
    elemental_scale: float = None
    n_species: int = 119


class PerElementRegressionShift:
    name = "per_element_regression_shift"
    parameters = ["energy_regularisation"]
    dtypes = [float]

    @staticmethod
    def compute(inputs, labels, shift_options) -> np.ndarray:
        log.info("Computing per element energy regression.")

        lambd = shift_options["energy_regularisation"]
        energies = labels["energy"]
        numbers = inputs["numbers"]
        system_sizes = inputs["n_atoms"]

        energies = np.array(energies)
        system_sizes = np.array(system_sizes)

        ds_energy = np.sum(energies)
        n_atoms_total = np.sum(system_sizes)

        mean_energy = ds_energy / n_atoms_total
        n_species = 119  # for simplicity, we assume any element could be in the dataset
        X = np.zeros(shape=(energies.shape[0], n_species))
        y = np.zeros(energies.shape[0])

        for i in range(energies.shape[0]):
            Z_counts = np.zeros(n_species)
            for z in set(numbers[i]):
                Z_counts[z] = np.count_nonzero(numbers[i] == z)
            X[i] = Z_counts

            E_sub_mean = energies[i] - mean_energy * numbers[i].shape[0]
            y[i] = E_sub_mean
        XTX = X.T @ X
        reg_term = lambd * np.eye(XTX.shape[0])

        result = np.linalg.lstsq(XTX + reg_term, X.T @ y, rcond=-1)
        elemental_energies_shift = result[0]
        elemental_energies_shift += mean_energy

        return elemental_energies_shift


class IsolatedAtomEnergyShift:
    name = "isolated_atom_energy_shift"
    parameters = ["E0s"]
    dtypes = [dict[int, float]]

    @staticmethod
    def compute(inputs, labels, shift_options):
        shift_options = shift_options["E0s"]
        n_species = 119
        elemental_energies_shift = np.zeros(n_species)
        for k, v in shift_options.items():
            elemental_energies_shift[k] = v

        return elemental_energies_shift


class MeanEnergyShift:
    name = "mean_atom_energy_shift"
    parameters = []
    dtypes = []

    @staticmethod
    def compute(inputs, labels, shift_options):
        energies = labels["energy"]
        system_sizes = inputs["n_atoms"]

        energies = np.array(energies)
        system_sizes = np.array(system_sizes)

        ds_energy = np.sum(energies)
        n_atoms_total = np.sum(system_sizes)

        mean_energy = ds_energy / n_atoms_total
        return mean_energy


class MeanEnergyRMSScale:
    name = "mean_energy_rms_scale"
    parameters = []
    dtypes = []

    @staticmethod
    def compute(inputs, labels, scale_options):
        # log.info("Computing per element energy regression.")
        energies = labels["energy"]
        numbers = inputs["numbers"]
        system_sizes = inputs["n_atoms"]

        energies = np.array(energies)
        system_sizes = np.array(system_sizes)

        ds_energy = np.sum(energies)
        n_atoms_total = np.sum(system_sizes)

        mean_energy = ds_energy / n_atoms_total

        mean_err_sse = 0.0

        for i in range(energies.shape[0]):
            E_sub_mean = energies[i] - mean_energy * numbers[i].shape[0]
            mean_err_sse += E_sub_mean**2 / numbers[i].shape[0]

        energy_std = np.sqrt(mean_err_sse / n_atoms_total)
        return energy_std


class PerElementForceRMSScale:
    name = "per_element_force_rms_scale"
    parameters = []
    dtypes = []

    @staticmethod
    def compute(inputs, labels, scale_options):
        n_species = 119

        forces = np.concatenate(labels["forces"], axis=0)
        numbers = np.concatenate(inputs["numbers"], axis=0)

        elements = np.unique(numbers)

        element_scale = np.ones(n_species)

        for element in elements:
            element_forces = forces[numbers == element]
            force_rms = np.sqrt(np.mean(np.linalg.norm(element_forces, 2, axis=1)))
            element_scale[element] = force_rms

        return element_scale


class GlobalCustomScale:
    name = "global_custom_scale"
    parameters = ["factor"]
    dtypes = [float]

    @staticmethod
    def compute(inputs, labels, scale_options):
        element_scale = scale_options["factor"]
        return element_scale


class PerElementCustomScale:
    name = "per_element_custom_scale"
    parameters = ["factors"]
    dtypes = [dict[int, float]]

    @staticmethod
    def compute(inputs, labels, scale_options):
        n_species = 119
        element_scale = np.ones(n_species)
        for k, v in scale_options["factors"].items():
            element_scale[k] = v

        return element_scale


shift_method_list = [PerElementRegressionShift, IsolatedAtomEnergyShift, MeanEnergyShift]
scale_method_list = [
    MeanEnergyRMSScale,
    PerElementForceRMSScale,
    GlobalCustomScale,
    PerElementCustomScale,
]


def compute_scale_shift_parameters(
    inputs, labels, shift_method, scale_method, shift_options, scale_options
):
    shift_methods = {method.name: method for method in shift_method_list}
    scale_methods = {method.name: method for method in scale_method_list}

    if shift_method not in shift_methods.keys():
        raise KeyError(
            f"The shift method '{shift_method}' is not among the implemented methods."
            f" Choose from {shift_method.keys()}"
        )
    if scale_method not in scale_methods.keys():
        raise KeyError(
            f"The scale method '{scale_method}' is not among the implemented methods."
            f" Choose from {scale_method.keys()}"
        )

    shift_method = shift_methods[shift_method]
    scale_method = scale_methods[scale_method]

    shift_parameters = shift_method.compute(inputs, labels, shift_options)
    scale_parameters = scale_method.compute(inputs, labels, scale_options)

    ds_stats = DatasetStats(shift_parameters, scale_parameters)
    return ds_stats



import os
import random
from typing import List

import jax
import numpy as np
import pytest
import yaml
from ase import Atoms
from ase.calculators.emt import EMT
from ase.calculators.singlepoint import SinglePointCalculator

from apax.config.train_config import Config
from apax.train.run import run
from apax.utils.datasets import download_md22_stachyose
from apax.utils.helpers import mod_config
from apax.utils.random import seed_py_np_tf


@pytest.fixture(autouse=True)
def set_radom_seeds():
    seed_py_np_tf()


def create_cell(a: float, lattice: str) -> np.ndarray:
    free_cell = (0.0, 0.0, 0.0)
    sc_cell = (a, a, a)
    fcc_cell = [(0, a, a), (a, 0, a), (a, a, 0)]
    bcc_cell = [(-a, a, a), (a, -a, a), (a, a, -a)]

    cells = {
        "free": free_cell,
        "sc": sc_cell,
        "fcc": fcc_cell,
        "bcc": bcc_cell,
    }

    return cells[lattice]


def create_example_atoms(num_atoms: int, pbc: bool = False, calc_results=[]):
    numbers = np.random.randint(1, 119, size=num_atoms)
    cell_const = np.random.uniform(low=10.0, high=12.0)
    positions = np.random.uniform(low=0.0, high=cell_const, size=(num_atoms, 3))

    additional_data = {}
    additional_data["pbc"] = pbc
    lattice = random.choice(["sc", "fcc", "bcc"])
    if pbc:
        additional_data["cell"] = create_cell(cell_const, lattice)
    else:
        additional_data["cell"] = [0.0, 0.0, 0.0]

    result_shapes = {
        "energy": (np.random.rand() - 5.0) * 10_000,
        "forces": np.random.uniform(low=-1.0, high=1.0, size=(num_atoms, 3)),
        # "stress": np.random.uniform(low=-1.0, high=1.0, size=(3, 3)),
        # "dipole": np.random.randn(3),
        # "charge": np.random.randint(-3, 4),
        # "ma_tensors": np.random.uniform(low=-1.0, high=1.0, size=(3, 3)),
    }

    atoms = Atoms(numbers=numbers, positions=positions, **additional_data)
    if calc_results:
        results = {}
        for key in calc_results:
            results[key] = result_shapes[key]

        atoms.calc = SinglePointCalculator(atoms, **results)
    return atoms


@pytest.fixture()
def example_atoms_list(num_data: int, pbc: bool, calc_results: List[str]) -> List[Atoms]:
    atoms_list = []

    for _ in range(num_data):
        num_atoms = np.random.randint(10, 15)
        atoms = create_example_atoms(num_atoms, pbc, calc_results)

        atoms_list.append(atoms)

    return atoms_list


@pytest.fixture()
def example_dataset(num_data: int) -> Atoms:
    atoms_list = []

    p2 = np.random.uniform(low=1.0, high=1.5, size=(num_data,))
    for i in range(num_data):
        positions = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, p2[i]]])

        additional_data = {}
        additional_data["cell"] = [0.0, 0.0, 0.0]

        atoms = Atoms("H2", positions=positions, **additional_data)
        atoms.calc = EMT()
        atoms.get_potential_energy()
        atoms.get_forces()

        atoms_list.append(atoms)

    return atoms_list


@pytest.fixture()
def get_tmp_path(tmp_path_factory):
    test_path = tmp_path_factory.mktemp("apax_tests")
    return test_path


@pytest.fixture(scope="session")
def tmp_data_path(tmp_path_factory):
    test_path = tmp_path_factory.mktemp("data")
    return test_path


@pytest.fixture(scope="session")
def get_md22_stachyose(tmp_data_path):
    file_path = download_md22_stachyose(tmp_data_path)
    return file_path


@pytest.fixture()
def get_sample_input():
    positions = np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ]
    )
    atomic_numbers = np.array([1, 1, 8])
    box = np.diag(np.zeros(3))
    offsets = np.full([3, 3], 0)
    idx = np.array([[1, 2, 0, 2, 0, 1], [0, 0, 1, 1, 2, 2]])

    inputs = (positions, atomic_numbers, idx, box, offsets)
    return inputs, box


def initialize_model(model_config, sample_input):
    Builder = model_config.model.get_builder()
    builder = Builder(model_config.model.model_dump())
    model = builder.build_energy_derivative_model()
    rng_key = jax.random.PRNGKey(model_config.seed)
    params = model.init(rng_key, *sample_input)

    return model, params


def load_and_dump_config(config_path, dump_path):
    with open(config_path.as_posix(), "r") as stream:
        model_config_dict = yaml.safe_load(stream)

    model_config_dict["data"]["directory"] = dump_path.as_posix()

    model_config = Config.model_validate(model_config_dict)
    os.makedirs(model_config.data.model_version_path, exist_ok=True)
    model_config.dump_config(model_config.data.model_version_path)
    return model_config


def load_config_and_run_training(config_path, updated_config):
    config_dict = mod_config(config_path, updated_config)
    run(config_dict)
    return config_dict



import mlflow


class MLFlowLogger:
    def __init__(self, experiment=None, run_name=None):
        """
        Initialize the MLFlow logger.

        :param log_dir: The directory where MLFlow logs will be stored.
        :param experiment_name: Name of the experiment in MLFlow.
        :param run_name: Name of the run to be logged in MLFlow.
        """
        mlflow.login()
        mlflow.set_experiment(experiment)

        # Start a new MLFlow run
        self.run = mlflow.start_run(run_name=run_name)

    def set_model(self, *args, **kwargs):
        pass

    def on_epoch_begin(self, epoch, logs=None):
        pass

    def on_train_batch_begin(self, batch, logs=None):
        pass

    def on_train_batch_end(self, batch, logs=None):
        pass

    def on_train_begin(self, logs=None):
        pass

    def on_epoch_end(self, epoch, logs=None):
        """Log metrics at the end of each epoch."""
        if logs is None:
            return
        mlflow.log_metrics(logs, step=epoch, synchronous=False)

    def on_train_end(self, logs=None):
        """
        End the current MLFlow run.
        """
        mlflow.end_run()



import os
from typing import Literal, Union

import yaml
from pydantic import BaseModel, Field, NonNegativeInt, PositiveFloat, PositiveInt
from typing_extensions import Annotated

from apax.utils.helpers import APAX_PROPERTIES


class ConstantTempSchedule(BaseModel, extra="forbid"):
    """Constant temperature schedule.

    Attributes
    ----------
    name: str
        Identifier of the temperature schedule.
    T0 : PositiveFloat, default = 298.15
        Initial temperature in Kelvin (K).
    """

    name: Literal["constant"] = "constant"
    T0: PositiveFloat = 298.15  # K

    def get_schedule(self):
        from apax.md.schedules import ConstantTSchedule

        return ConstantTSchedule(self.T0)


class PiecewiseLinearTempSchedule(ConstantTempSchedule, extra="forbid"):
    """Piecewise linear temperature schedule.
    Temperature is linearly interpolated between T0 and the supplied
    values at the specified time steps.

    Attributes
    ----------
    temperatures: list[PositiveFloat]
        List of temperatures to interpolate between.
    durations: list[PositiveInt]
        Duration (in time steps) of the interpolation between two
        subsequent values of `temperatures`.
    """

    name: Literal["piecewise"] = "piecewise"
    temperatures: list[PositiveFloat]
    durations: list[PositiveInt]

    def get_schedule(self):
        from apax.md.schedules import PieceWiseLinearTSchedule

        schedule = PieceWiseLinearTSchedule(
            self.T0,
            self.temperatures,
            self.durations,
        )
        return schedule


class OscillatingRampTempSchedule(ConstantTempSchedule, extra="forbid"):
    """Combination of a linear interpolation between T0 and Tend and a temperature oscillation.
    Mostly for sampling purposes.

    Attributes
    ----------
    Tend: PositiveFloat
        Final temperature in Kelvin.
    amplitude: PositiveFloat
        Amplitude of temperature oscilaltions.
    num_oscillations: PositiveInt
        Number of oscillations to occur during the simulation.
    total_steps: PositiveInt
        Total steps of the schedule. Afterwards, Tend will be kept.
    """

    name: Literal["oscillating_ramp"] = "oscillating_ramp"
    Tend: PositiveFloat
    amplitude: PositiveFloat
    num_oscillations: PositiveInt
    total_steps: PositiveInt

    def get_schedule(self):
        from apax.md.schedules import OscillatingRampTSchedule

        schedule = OscillatingRampTSchedule(
            self.T0,
            self.Tend,
            self.amplitude,
            self.num_oscillations,
            self.total_steps,
        )
        return schedule


TemperatureSchedule = Union[
    ConstantTempSchedule, PiecewiseLinearTempSchedule, OscillatingRampTempSchedule
]


class NHCOptions(BaseModel, extra="forbid"):
    """
    Options for Nose-Hoover chain thermostat.

    Parameters
    ----------
    chain_length : PositiveInt, default = 3
        Number of thermostats in the chain.
    chain_steps : PositiveInt, default = 2
        Number of steps per chain.
    sy_steps : PositiveInt, default = 3
        Number of steps for Suzuki-Yoshida integration.
    tau : PositiveFloat, default = 100
        Relaxation time parameter.
    """

    chain_length: PositiveInt = 3
    chain_steps: PositiveInt = 2
    sy_steps: PositiveInt = 3
    tau: PositiveFloat = 100


class Integrator(BaseModel, extra="forbid"):
    """
    Molecular dynamics integrator options.

    Parameters
    ----------
    name : str
        Name of the ensemble.
    dt: PositiveFloat, default = 0.5
        Time step in femto seconds.
    temperature_schedule: TemperatureSchedule
        Temperature schedule to use throughout the simulation.
        For NVE, it is only used for velocity initialization and
        disregarded at subsequent steps.
    """

    name: str
    dt: PositiveFloat = 0.5  # fs
    temperature_schedule: TemperatureSchedule = Field(
        ConstantTempSchedule(name="constant", T0=298.15), discriminator="name"
    )


class NVEOptions(Integrator, extra="forbid"):
    """
    Options for NVE ensemble simulations.

    Attributes
    ----------
    name : Literal["nve"]
        Name of the ensemble.
    """

    name: Literal["nve"]


class NVTOptions(NVEOptions, extra="forbid"):
    """
    Options for NVT ensemble simulations.

    Parameters
    ----------
    name : Literal["nvt"]
        Name of the ensemble.
    thermostat_chain : NHCOptions, default = NHCOptions()
        Thermostat chain options.
    """

    name: Literal["nvt"]
    thermostat_chain: NHCOptions = NHCOptions()


class NPTOptions(NVTOptions, extra="forbid"):
    """
    Options for NPT ensemble simulations.

    Parameters
    ----------
    name : Literal["npt"]
        Name of the ensemble.
    pressure : PositiveFloat, default = 1.01325
        Pressure in bar.
    barostat_chain : NHCOptions, default = NHCOptions(tau=1000)
        Barostat chain options.
    """

    name: Literal["npt"]
    pressure: PositiveFloat = 1.01325  # bar
    barostat_chain: NHCOptions = NHCOptions(tau=1000)


class EnergyUncertaintyCheck(BaseModel, extra="forbid"):
    name: Literal["energy_uncertainty"] = "energy_uncertainty"
    threshold: PositiveFloat
    per_atom: bool = True


class ForcesUncertaintyCheck(BaseModel, extra="forbid"):
    name: Literal["forces_uncertainty"] = "forces_uncertainty"
    threshold: PositiveFloat


class ReflectionCheck(BaseModel, extra="forbid"):
    name: Literal["reflection"] = "reflection"
    cutoff_plane_height: float


DynamicsCheck = Annotated[
    Union[EnergyUncertaintyCheck, ForcesUncertaintyCheck, ReflectionCheck],
    Field(discriminator="name"),
]


class FixAtomsConstraint(BaseModel, extra="forbid"):
    name: Literal["fixatoms"] = "fixatoms"
    indices: list[int]


class FixLayerConstraint(BaseModel, extra="forbid"):
    name: Literal["fixlayer"] = "fixlayer"
    upper_limit: float
    lower_limit: float


Constraint = Annotated[
    Union[FixAtomsConstraint, FixLayerConstraint], Field(discriminator="name")
]


class H5MDOptions(BaseModel, extra="forbid"):
    """
    Options for configuring H5MD (Hierarchical Data Format for Molecular Data) file output.

    Attributes
    ----------
    compression: str | None
        | Compression algorithm to use for the HDF5 file (e.g., "gzip"). Defaults to "gzip".
    compression_opts: int
        | Compression level or options for the chosen algorithm. Defaults to 4.
    store: Literal["time", "linear"]
        | Storage mode for the data. Can be "time" or "linear". Defaults to "time".
    author: str
        | Name of the author of the file. Defaults to "N/A".
    author_email: str
        | Email address of the author. Defaults to "N/A".
    """

    compression: str | None = "gzip"
    compression_opts: int = 4
    store: Literal["time", "linear"] = "time"
    author: str = "N/A"
    author_email: str = "N/A"


class MDConfig(BaseModel, frozen=True, extra="forbid"):
    """
    Configuration for a NHC molecular dynamics simulation.
    Full config :ref:`here <md_config>`:

    Parameters
    ----------
    seed : int, default = 1
        | Random seed for momentum initialization.
    ensemle :
        | Options for integrating the EoM and which ensemble to use.
    dt : float, default = 0.5
        | Time step in fs.
    duration : float, required
        | Total simulation time in fs.
    n_inner : int, default = 500
        | Number of compiled simulation steps (i.e. number of iterations of the
        | `jax.lax.fori_loop` loop). Also determines atoms buffer size.
    sampling_rate : int, default = 10
        | Interval between saving frames.
    buffer_size : int, default = 2500
        | Number of collected frames to be dumped at once.
    dr_threshold : float, default = 0.5
        | Skin of the neighborlist.
    extra_capacity : int, default = 0
        | JaxMD allocates a maximal number of neighbors. This argument lets you add
        | additional capacity to avoid recompilation. The default is usually fine.

    dynamics_checks: list[DynamicsCheck]
        | List of termination criteria. Currently energy and force uncertainty
        | are available
    properties: list[str]
        | Whitelist of properties to be saved in the trajectory.
        | This does not effect what the model will calculate, e.g..
        | an ensemble will still calculate uncertainties.
    initial_structure : str, required
        | Path to the starting structure of the simulation.
    sim_dir : str, default = "."
        | Directory where simulation file will be stored.
    traj_name : str, default = "md.h5"
        | Name of the trajectory file.
    restart : bool, default = True
        | Whether the simulation should restart from the latest configuration in
        | `traj_name`.
    wrapped : bool, default = True
        | Whether the atoms in the simulation should wrapped back into the box.
    checkpoint_interval : int, default = 50_000
        | Number of time steps between saving full simulation state checkpoints.
        | These will be loaded with the `restart` option.
    disable_pbar : bool, False
        | Disables the MD progressbar.
    """

    seed: int = 1

    # https://docs.pydantic.dev/latest/usage/types/unions/#discriminated-unions-aka-tagged-unions
    ensemble: Union[NVEOptions, NVTOptions, NPTOptions] = Field(
        NVTOptions(name="nvt"), discriminator="name"
    )

    duration: PositiveFloat
    n_inner: PositiveInt = 500
    sampling_rate: PositiveInt = 10
    buffer_size: PositiveInt = 2500
    dr_threshold: PositiveFloat = 0.5
    extra_capacity: NonNegativeInt = 0

    dynamics_checks: list[DynamicsCheck] = []
    constraints: list[Constraint] = []

    properties: list[str] = APAX_PROPERTIES
    h5md_options: H5MDOptions = H5MDOptions()

    initial_structure: str
    load_momenta: bool = False
    sim_dir: str = "."
    traj_name: str = "md.h5"
    restart: bool = True
    wrapped: bool = True
    checkpoint_interval: int = 50_000
    disable_pbar: bool = False

    def dump_config(self):
        """
        Writes the current config file to the MD directory.

        """
        with open(os.path.join(self.sim_dir, "md_config.yaml"), "w") as conf:
            yaml.dump(self.model_dump(), conf, default_flow_style=False)



import logging

import numpy as np

from apax.config import ModelConfig
from apax.layers.descriptor import (
    EquivMPRepresentation,
    GaussianMomentDescriptor,
    So3kratesRepresentation,
)
from apax.layers.descriptor.basis_functions import (
    BesselBasis,
    GaussianBasis,
    RadialFunction,
)
from apax.layers.empirical import all_corrections
from apax.layers.properties import PropertyHead
from apax.layers.readout import AtomisticReadout
from apax.layers.scaling import PerElementScaleShift
from apax.nn.models import (
    EnergyDerivativeModel,
    EnergyModel,
    FeatureModel,
    ShallowEnsembleModel,
)

log = logging.getLogger(__name__)


class ModelBuilder:
    def __init__(self, model_config: ModelConfig, n_species: int = 119):
        self.config = model_config
        self.n_species = n_species

    def build_basis_function(self):
        basis_config = self.config["basis"]
        name = basis_config["name"]

        if name == "gaussian":
            basis_fn = GaussianBasis(
                n_basis=basis_config["n_basis"],
                r_min=basis_config["r_min"],
                r_max=basis_config["r_max"],
                dtype=self.config["descriptor_dtype"],
            )
        elif name == "bessel":
            basis_fn = BesselBasis(
                n_basis=basis_config["n_basis"],
                r_max=basis_config["r_max"],
                dtype=self.config["descriptor_dtype"],
            )
        else:
            raise ValueError("unknown basis requested")
        return basis_fn

    def build_radial_function(self):
        basis_fn = self.build_basis_function()

        if self.config["basis"]["name"] == "gaussian":
            use_embed_norm = True
            one_sided_dist = False
        else:
            use_embed_norm = False
            one_sided_dist = True

        radial_fn = RadialFunction(
            n_radial=self.config["n_radial"],
            basis_fn=basis_fn,
            n_species=self.n_species,
            emb_init=self.config["emb_init"],
            use_embed_norm=use_embed_norm,
            one_sided_dist=one_sided_dist,
            dtype=self.config["descriptor_dtype"],
        )
        return radial_fn

    def build_descriptor(
        self,
        apply_mask,
    ):
        raise NotImplementedError("use a subclass to facilitate this")

    def build_readout(
        self, head_config, is_feature_fn=False, only_use_n_layers: None | int = None
    ):
        has_ensemble = "ensemble" in head_config.keys() and head_config["ensemble"]
        if has_ensemble and head_config["ensemble"]["kind"] == "shallow":
            n_shallow_ensemble = head_config["ensemble"]["n_members"]
        elif "n_shallow_members" in head_config.keys():
            n_shallow_ensemble = head_config["n_shallow_members"]
        else:
            n_shallow_ensemble = 0

        if "readout_dtype" in head_config:
            dtype = head_config["readout_dtype"]
        elif "dtype" in head_config:
            dtype = head_config["dtype"]
        else:
            raise KeyError("No dtype specified in config")

        nn_layers = head_config["nn"]
        if only_use_n_layers is not None:
            nn_layers = nn_layers[:only_use_n_layers]
            if len(nn_layers) == 0:
                return None

        readout = AtomisticReadout(
            units=nn_layers,
            b_init=head_config["b_init"],
            w_init=head_config["w_init"],
            use_ntk=head_config["use_ntk"],
            is_feature_fn=is_feature_fn,
            n_shallow_ensemble=n_shallow_ensemble,
            dtype=dtype,
        )
        return readout

    def build_scale_shift(self, scale, shift):
        scale_shift = PerElementScaleShift(
            n_species=self.n_species,
            scale=scale,
            shift=shift,
            dtype=self.config["scale_shift_dtype"],
        )
        return scale_shift

    def build_property_heads(self, apply_mask: bool = True):
        property_heads = []
        for head in self.config["property_heads"]:
            readout = self.build_readout(head)
            phead = PropertyHead(
                pname=head["name"],
                aggregation=head["aggregation"],
                mode=head["mode"],
                readout=readout,
                apply_mask=apply_mask,
            )
            property_heads.append(phead)
        return property_heads

    def build_corrections(self, apply_mask: bool = True):
        corrections = []
        for correction in self.config["empirical_corrections"]:
            correction = correction.copy()
            name = correction.pop("name")
            Correction = all_corrections[name]
            corr = Correction(
                **correction,
                apply_mask=apply_mask,
            )
            corrections.append(corr)

        return corrections

    def build_energy_model(
        self,
        scale=1.0,
        shift=0.0,
        apply_mask=True,
        init_box: np.array = np.array([0.0, 0.0, 0.0]),
        inference_disp_fn=None,
    ):
        log.debug("Building atomistic model")

        descriptor = self.build_descriptor(apply_mask)
        readout = self.build_readout(self.config)
        scale_shift = self.build_scale_shift(scale, shift)

        property_heads = self.build_property_heads(apply_mask=apply_mask)
        corrections = self.build_corrections(apply_mask=apply_mask)

        model = EnergyModel(
            representation=descriptor,
            readout=readout,
            scale_shift=scale_shift,
            property_heads=property_heads,
            corrections=corrections,
            init_box=init_box,
            inference_disp_fn=inference_disp_fn,
        )
        return model

    def build_energy_derivative_model(
        self,
        scale=1.0,
        shift=0.0,
        apply_mask=True,
        init_box: np.array = np.array([0.0, 0.0, 0.0]),
        inference_disp_fn=None,
    ):
        energy_model = self.build_energy_model(
            scale,
            shift,
            apply_mask,
            init_box=init_box,
            inference_disp_fn=inference_disp_fn,
        )
        if (
            self.config["ensemble"]
            and self.config["ensemble"]["kind"] == "shallow"
            and self.config["ensemble"]["n_members"] > 1
        ):
            log.info("Building ShallowEnsemble model")
            model = ShallowEnsembleModel(
                energy_model,
                calc_stress=self.config["calc_stress"],
                force_variance=self.config["ensemble"]["force_variance"],
                chunk_size=self.config["ensemble"]["chunk_size"],
            )
        else:
            log.info("Building Standard model")
            model = EnergyDerivativeModel(
                energy_model,
                calc_stress=self.config["calc_stress"],
            )
        return model

    def build_feature_model(
        self,
        only_use_n_layers=None,
        apply_mask=True,
        init_box: np.array = np.array([0.0, 0.0, 0.0]),
        inference_disp_fn=None,
    ):
        log.info("Building feature model")
        descriptor = self.build_descriptor(apply_mask)
        readout = self.build_readout(
            self.config, is_feature_fn=True, only_use_n_layers=only_use_n_layers
        )

        model = FeatureModel(
            descriptor,
            readout,
            should_average=True,
            init_box=init_box,
            inference_disp_fn=inference_disp_fn,
            mask_atoms=True,
        )
        return model


class GMNNBuilder(ModelBuilder):
    def build_descriptor(
        self,
        apply_mask,
    ):
        radial_fn = self.build_radial_function()
        descriptor = GaussianMomentDescriptor(
            radial_fn=radial_fn,
            n_contr=self.config["n_contr"],
            dtype=self.config["descriptor_dtype"],
            apply_mask=apply_mask,
        )
        return descriptor


class EquivMPBuilder(ModelBuilder):
    def build_descriptor(
        self,
        apply_mask,
    ):
        descriptor = EquivMPRepresentation(
            features=self.config["features"],
            max_degree=self.config["max_degree"],
            num_iterations=self.config["num_iterations"],
            basis_fn=self.build_basis_function(),
            dtype=self.config["descriptor_dtype"],
            apply_mask=apply_mask,
        )
        return descriptor


class So3kratesBuilder(ModelBuilder):
    def build_descriptor(
        self,
        apply_mask,
    ):
        descriptor = So3kratesRepresentation(
            basis_fn=self.build_basis_function(),
            num_layers=self.config["num_layers"],
            max_degree=self.config["max_degree"],
            num_features=self.config["num_features"],
            num_heads=self.config["num_heads"],
            use_layer_norm_1=self.config["use_layer_norm_1"],
            use_layer_norm_2=self.config["use_layer_norm_2"],
            use_layer_norm_final=self.config["use_layer_norm_final"],
            activation=self.config["activation"],
            cutoff_fn=self.config["cutoff_fn"],
            transform_input_features=self.config["transform_input_features"],
            dtype=self.config["descriptor_dtype"],
        )
        return descriptor



import jax
import jax.numpy as jnp
import numpy as np

from apax.layers.descriptor import GaussianMomentDescriptor
from apax.layers.scaling import PerElementScaleShift
from apax.nn.models import EnergyDerivativeModel, EnergyModel


def test_apax_variable_size():
    R = np.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
        ]
    )

    Z = np.array([1, 2, 2])

    idx = np.array(
        [
            [1, 2, 0, 2, 0, 1],
            [0, 0, 1, 1, 2, 2],
        ]
    )
    offsets = jnp.full([6, 3], 0)
    box = np.array([0, 0, 0])

    R_padded = np.concatenate([R, np.zeros((1, 3))], axis=0)
    Z_padded = np.concatenate([Z, [0]])
    idx_padded = np.concatenate([idx, [[0], [0]]], axis=1)

    shift = jnp.array([0.0, 100.0, 200.0])
    scale = jnp.array([1.0, 1.2, 1.6])[..., None]

    model = EnergyDerivativeModel(
        EnergyModel(
            representation=GaussianMomentDescriptor(apply_mask=False),
            scale_shift=PerElementScaleShift(scale=scale, shift=shift),
            mask_atoms=False,
        )
    )
    model_padded = EnergyDerivativeModel(
        EnergyModel(
            representation=GaussianMomentDescriptor(apply_mask=True),
            scale_shift=PerElementScaleShift(scale=scale, shift=shift),
            mask_atoms=True,
        )
    )

    rng_key = jax.random.PRNGKey(1)

    params = model.init(rng_key, R, Z, idx, box, offsets)

    results = model.apply(params, R, Z, idx, box, offsets)
    results_padded = model_padded.apply(
        params, R_padded, Z_padded, idx_padded, box, offsets
    )

    assert (results["energy"] - results_padded["energy"]) < 1e-6
    assert np.all(results["forces"] - results_padded["forces"][:-1, :] < 1e-6)  # 1e-6


def test_energy_model():
    key = jax.random.PRNGKey(0)

    R = np.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
        ]
    )

    Z = np.array([1, 2, 2])

    idx = np.array(
        [
            [1, 2, 0, 2, 0, 1],
            [0, 0, 1, 1, 2, 2],
        ]
    )
    offsets = jnp.full([6, 3], 0)
    box = np.array([0.0, 0.0, 0.0])

    model = EnergyModel()

    params = model.init(key, R, Z, idx, box, offsets)
    energy, properties = model.apply(params, R, Z, idx, box, offsets)

    assert energy.shape == ()
    assert isinstance(properties, dict)


def test_energy_force_model():
    key = jax.random.PRNGKey(0)

    R = np.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
        ]
    )

    Z = np.array([1, 2, 2])

    idx = np.array(
        [
            [1, 2, 0, 2, 0, 1],
            [0, 0, 1, 1, 2, 2],
        ]
    )
    offsets = jnp.full([6, 3], 0)

    box = np.array([0.0, 0.0, 0.0])

    model = EnergyDerivativeModel()

    params = model.init(key, R, Z, idx, box, offsets)
    result = model.apply(params, R, Z, idx, box, offsets)

    assert result["energy"].shape == ()
    assert result["forces"].shape == (3, 3)



import logging
import time
from functools import partial
from pathlib import Path

import jax
import jax.numpy as jnp
import numpy as np
from ase import units
from ase.io import read
from flax.training import checkpoints
from jax import tree_util
from jax.experimental import io_callback
from tqdm import trange
from tqdm.contrib.logging import logging_redirect_tqdm

from apax.config import Config, MDConfig, parse_config
from apax.config.md_config import Integrator
from apax.md.ase_calc import make_ensemble, maybe_vmap
from apax.md.constraints import Constraint, ConstraintBase
from apax.md.dynamics_checks import DynamicsCheckBase, DynamicsChecks
from apax.md.io import H5TrajHandler, TrajHandler, truncate_trajectory_to_checkpoint
from apax.md.md_checkpoint import load_md_state
from apax.md.sim_utils import SimulationFunctions, System
from apax.train.checkpoints import (
    canonicalize_energy_model_parameters,
    restore_parameters,
)
from apax.train.run import setup_logging
from apax.utils.jax_md_reduced import partition, quantity, simulate, space
from apax.utils.transform import make_energy_only_model

log = logging.getLogger(__name__)


def create_energy_fn(model, params, numbers, n_models, shallow=False):
    def full_ensemble(params, R, Z, neighbor, box, offsets, perturbation=None):
        vmodel = jax.vmap(model, (0, None, None, None, None, None, None), 0)
        energies, _ = vmodel(params, R, Z, neighbor, box, offsets, perturbation)
        energy = jnp.mean(energies)
        return energy

    def shallow_ensemble(params, R, Z, neighbor, box, offsets, perturbation=None):
        energies, _ = model(params, R, Z, neighbor, box, offsets, perturbation)
        energy = jnp.mean(energies)
        return energy

    if n_models > 1:
        if shallow:
            energy_fn = shallow_ensemble
        else:
            energy_fn = full_ensemble
    else:
        energy_fn = make_energy_only_model(model)

    energy_fn = partial(
        energy_fn,
        params,
        Z=numbers,
        offsets=jnp.array([0.0, 0.0, 0.0]),
    )

    return energy_fn


def heights_of_box_sides(box):
    heights = []

    for i in range(len(box)):
        for j in range(i + 1, len(box)):
            area = np.linalg.norm(np.cross(box[i], box[j]))
            height = area / np.linalg.norm(box[i])
            heights.append(height)
            height = area / np.linalg.norm(box[j])
            heights.append(height)

    return np.array(heights)


def nbr_update_options_default(state):
    return {}


def nbr_update_options_npt(state):
    box = simulate.npt_box(state)
    return {"box": box}


def get_ensemble(ensemble: Integrator, sim_fns):
    energy, shift = sim_fns.energy_fn, sim_fns.shift_fn

    dt = ensemble.dt * units.fs
    nbr_options = nbr_update_options_default

    kT = ensemble.temperature_schedule.get_schedule()
    if ensemble.name == "nve":
        init_fn, apply_fn = simulate.nve(energy, shift, kT(0), dt)
    elif ensemble.name == "nvt":
        thermostat_chain = dict(ensemble.thermostat_chain)
        thermostat_chain["tau"] *= dt

        init_fn, apply_fn = simulate.nvt_nose_hoover(energy, shift, dt, kT(0))

    elif ensemble.name == "npt":
        pressure = ensemble.pressure * units.bar
        thermostat_chain = dict(ensemble.thermostat_chain)
        barostat_chain = dict(ensemble.barostat_chain)
        thermostat_chain["tau"] *= dt
        barostat_chain["tau"] *= dt

        init_fn, apply_fn = simulate.npt_nose_hoover(
            energy,
            shift,
            dt,
            pressure,
            kT(0),
            thermostat_kwargs=thermostat_chain,
            barostat_kwargs=barostat_chain,
        )
        nbr_options = nbr_update_options_npt
    else:
        raise NotImplementedError(
            "Only the NVE and Nose Hoover NVT/NPT thermostats are currently interfaced."
        )

    return init_fn, apply_fn, kT, nbr_options


def handle_checkpoints(state, step, system, load_momenta, ckpt_dir, should_load_ckpt):
    if load_momenta and not should_load_ckpt:
        log.info("loading momenta from starting configuration")
        state = state.set(momentum=system.momenta)

    elif should_load_ckpt:
        state, step = load_md_state(state, ckpt_dir)
    return state, step


def create_evaluation_functions(aux_fn, positions, Z, neighbor, box, dynamics_checks):
    offsets = jnp.zeros((neighbor.idx.shape[1], 3))

    def on_eval(positions, neighbor, box):
        predictions = aux_fn(positions, Z, neighbor, box, offsets)
        all_checks_passed = True

        for check in dynamics_checks:
            check_passed = check.check(predictions, positions, box)
            all_checks_passed = all_checks_passed & check_passed
        return predictions, all_checks_passed

    predictions = aux_fn(positions, Z, neighbor, box, offsets)
    dummpy_preds = tree_util.tree_map(lambda x: jnp.zeros_like(x), predictions)

    def no_eval(positions, neighbor, box):
        predictions = dummpy_preds
        all_checks_passed = True
        return predictions, all_checks_passed

    return on_eval, no_eval


def create_constraint_function(constraints: list[ConstraintBase], state, system):
    all_constraints = [c.create(state, system) for c in constraints]

    def apply_constraints(state):
        for constraint in all_constraints:
            state = constraint(state)

        return state

    return apply_constraints


def run_sim(
    system: System,
    sim_fns: SimulationFunctions,
    ensemble,
    sim_dir: Path,
    n_steps: int,
    n_inner: int,
    extra_capacity: int,
    rng_key: int,
    traj_handler: TrajHandler,
    load_momenta: bool = False,
    restart: bool = True,
    checkpoint_interval: int = 50_000,
    dynamics_checks: list[DynamicsCheckBase] = [],
    constraints: list[ConstraintBase] = [],
    disable_pbar: bool = False,
):
    """
    Performs NVT MD.

    Parameters
    ----------
    ensemble :
        Thermodynamic ensemble.
    n_steps : int
        Total time steps.
    n_inner : int
        JIT compiled inner loop. Also determines atoms buffer size.
    extra_capacity : int
        Extra capacity for the neighborlist.
    rng_key : int
        RNG key used to initialize the simulation.
    restart : bool, default = True
        Whether a checkpoint should be loaded. No implemented yet.
    checkpoint_interval : int, default = 50_000
        Number of time steps between saving
        full simulation state checkpoints.
    sim_dir : Path
        Directory where the trajectory and simulation checkpoints will be saved.
    """
    neighbor_fn = sim_fns.neighbor_fn
    ckpt_dir = sim_dir / "ckpts"
    ckpt_dir.mkdir(exist_ok=True)

    log.info("initializing simulation")
    init_fn, apply_fn, kT, nbr_options = get_ensemble(ensemble, sim_fns)

    neighbor = sim_fns.neighbor_fn.allocate(
        system.positions, extra_capacity=extra_capacity
    )

    state = init_fn(
        rng_key,
        system.positions,
        box=system.box,
        mass=system.masses,
        neighbor=neighbor,
    )

    step = 0
    ckpts_exist = any(True for p in ckpt_dir.rglob("*") if "checkpoint" in p.stem)
    should_load_ckpt = restart and ckpts_exist
    state, step = handle_checkpoints(
        state, step, system, load_momenta, ckpt_dir, should_load_ckpt
    )
    if should_load_ckpt:
        length = step * n_inner
        truncate_trajectory_to_checkpoint(traj_handler.traj_path, length)

    async_manager = checkpoints.AsyncManager()

    n_outer = int(np.ceil(n_steps / n_inner))
    pbar_update_freq = int(np.ceil(500 / n_inner))
    pbar_increment = n_inner * pbar_update_freq

    sampling_rate = traj_handler.sampling_rate
    on_eval, no_eval = create_evaluation_functions(
        sim_fns.auxiliary_fn,
        state.position,
        system.atomic_numbers,
        neighbor,
        system.box,
        dynamics_checks,
    )

    apply_constraints = create_constraint_function(
        constraints,
        state,
        system,
    )

    @jax.jit
    def sim(state, outer_step, neighbor):  # TODO make more modular
        def body_fn(i, state):
            state, outer_step, neighbor, all_checks_passed = state
            step = i + outer_step * n_inner

            apply_fn_kwargs = {}
            if isinstance(state, simulate.NPTNoseHooverState):
                box = state.box
            else:
                box = system.box
                apply_fn_kwargs = {"box": box}

            apply_fn_kwargs["kT"] = kT(step)  # Get current Temperature

            state = apply_fn(state, neighbor=neighbor, **apply_fn_kwargs)

            state = apply_constraints(state)

            nbr_kwargs = nbr_options(state)
            neighbor = neighbor.update(state.position, **nbr_kwargs)

            condition = step % sampling_rate == 0
            predictions, check_passed = jax.lax.cond(
                condition, on_eval, no_eval, state.position, neighbor, box
            )

            all_checks_passed = all_checks_passed & check_passed

            # maybe move this to on_eval
            io_callback(traj_handler.step, None, (state, predictions, nbr_kwargs))
            return state, outer_step, neighbor, all_checks_passed

        all_checks_passed = True
        state, outer_step, neighbor, all_checks_passed = jax.lax.fori_loop(
            0, n_inner, body_fn, (state, outer_step, neighbor, all_checks_passed)
        )
        current_temperature = (
            quantity.temperature(velocity=state.velocity, mass=state.mass) / units.kB
        )

        return state, neighbor, current_temperature, all_checks_passed

    start = time.time()
    total_sim_time = n_steps * ensemble.dt / 1000
    log.info("running simulation for %.1f ps", total_sim_time)
    initial_time = step * n_inner
    sim_pbar = trange(
        initial_time,
        n_steps,
        initial=initial_time,
        total=n_steps,
        desc="Simulation",
        ncols=100,
        disable=disable_pbar,
        leave=True,
    )
    while step < n_outer:
        new_state, neighbor, current_temperature, all_checks_passed = sim(
            state, step, neighbor
        )

        if np.any(np.isnan(state.position)) or np.any(np.isnan(state.velocity)):
            raise ValueError(
                f"NaN encountered, simulation aborted after {step + 1} steps."
            )

        if not all_checks_passed:
            with logging_redirect_tqdm():
                log.critical(f"One or more dynamics checks failed at step: {step + 1}")
            break

        if neighbor.did_buffer_overflow:
            with logging_redirect_tqdm():
                log.warn("step %d: neighbor list overflowed, reallocating.", step)
            traj_handler.reset_buffer()
            neighbor = neighbor_fn.allocate(
                state.position
            )  # TODO check that this actually works
        else:
            state = new_state
            step += 1

            if step % checkpoint_interval == 0:
                with logging_redirect_tqdm():
                    current_sim_time = step * n_inner * ensemble.dt / 1000
                    log.info(
                        f"saving checkpoint at {current_sim_time:.1f} ps - step: {step}"
                    )
                ckpt = {"state": state, "step": step}
                checkpoints.save_checkpoint(
                    ckpt_dir=ckpt_dir.resolve(),
                    target=ckpt,
                    step=step,
                    overwrite=True,
                    keep=2,
                    async_manager=async_manager,
                )

            if step % pbar_update_freq == 0:
                sim_pbar.set_postfix(T=f"{(current_temperature):.1f} K")  # set string
                sim_pbar.update(pbar_increment)

    # In case of mismatch update freq and n_steps, we can set it to 100% manually
    sim_pbar.update(n_steps - sim_pbar.n)
    sim_pbar.close()

    ckpt = {"state": state, "step": step}
    checkpoints.save_checkpoint(
        ckpt_dir=ckpt_dir.resolve(),
        target=ckpt,
        step=step,
        overwrite=True,
        keep=2,
        async_manager=async_manager,
    )
    traj_handler.write()
    traj_handler.close()
    end = time.time()
    elapsed_wall_time = end - start
    elapsed_sim_time = step * n_inner * ensemble.dt / 1000

    ps_per_s = elapsed_sim_time / elapsed_wall_time
    nanosec_per_day = ps_per_s / 1e3 * 60 * 60 * 24

    sec_per_step = elapsed_wall_time / n_steps
    n_atoms = system.positions.shape[0]
    musec_per_step_per_atom = sec_per_step * 1e6 / n_atoms

    log.info("simulation finished after: %.2f s", elapsed_wall_time)
    log.info(
        "performance summary: %.2f ns/day, %.2f mu s/step/atom",
        nanosec_per_day,
        musec_per_step_per_atom,
    )


def md_setup(model_config: Config, md_config: MDConfig):
    """
    Sets up the energy and neighborlist functions for an MD simulation,
    loads the initial structure.

    Parameters
    ----------
    model_config : Config
        Configuration of the model used as an interatomic potential.
    md_config : MDConfig
        configuration of the MD simulation.

    Returns
    -------
    R:
        Initial positions in Angstrom.
    atomic_numbers:
        Atomic numbers of the system.
    masses:
        Atomic masses in ASE units.
    box:
        Side length of the cubic box.
    energy_fn:
        Interatomic potential.
    neighbor_fn:
        Neighborlist function.
    shift_fn:
        Shift function for the integrator.
    """
    log.info("reading structure")
    atoms = read(md_config.initial_structure)
    system = System.from_atoms(atoms)

    r_max = model_config.model.basis.r_max
    log.info("initializing model")
    if np.all(system.box < 1e-6):
        frac_coords = False
        displacement_fn, shift_fn = space.free()
    else:
        frac_coords = True
        heights = heights_of_box_sides(system.box)

        if np.any(atoms.cell.lengths() / 2 < r_max):
            log.error(
                f"Cutoff radius is larger than half the box in at least one cell vector direction: "
                f"{r_max} > {np.min(atoms.cell.lengths()) / 2}. Cannot calculate correct neighbors."
            )
        if np.any(heights / 2 < r_max):
            log.error(
                f"Cutoff radius is larger than half the box in at least one cell vector direction: "
                f"{r_max} > {np.min(heights) / 2}. Cannot calculate correct neighbors."
            )

        displacement_fn, shift_fn = space.periodic_general(
            system.box,
            fractional_coordinates=frac_coords,
            wrapped=md_config.wrapped,
        )

    Builder = model_config.model.get_builder()
    builder = Builder(model_config.model.model_dump())
    energy_model = builder.build_energy_model(
        apply_mask=True, init_box=np.array(system.box), inference_disp_fn=displacement_fn
    )
    neighbor_fn = partition.neighbor_list(
        displacement_fn,
        system.box,
        r_max,
        md_config.dr_threshold,
        fractional_coordinates=frac_coords,
        format=partition.Sparse,
        disable_cell_list=True,
    )

    _, gradient_model_params = restore_parameters(model_config.data.model_version_path)
    params = canonicalize_energy_model_parameters(gradient_model_params)

    n_models = 1
    shallow = False
    if (
        "ensemble" in model_config.model.model_dump().keys()
        and model_config.model.ensemble is not None
        and model_config.model.ensemble.n_members > 1
    ):
        n_models = model_config.model.ensemble.n_members
        if model_config.model.ensemble.kind == "shallow":
            shallow = True

    energy_fn = create_energy_fn(
        energy_model.apply, params, system.atomic_numbers, n_models, shallow
    )

    auxiliary_fn = builder.build_energy_derivative_model(
        apply_mask=True, init_box=np.array(system.box), inference_disp_fn=displacement_fn
    ).apply

    if n_models > 1 and not shallow:
        auxiliary_fn = maybe_vmap(auxiliary_fn, gradient_model_params)
        auxiliary_fn = make_ensemble(auxiliary_fn)
    else:
        auxiliary_fn = partial(
            auxiliary_fn,
            gradient_model_params,
        )

    sim_fns = SimulationFunctions(energy_fn, auxiliary_fn, shift_fn, neighbor_fn)
    return system, sim_fns


def run_md(model_config: Config, md_config: MDConfig, log_level="error"):
    """
    Utiliy function to start NVT molecualr dynamics simulations from
    a previously trained model.

    Parameters
    ----------
    model_config : Config
        Configuration of the model used as an interatomic potential.
    md_config : MDConfig
        configuration of the MD simulation.
    """

    model_config = parse_config(model_config)
    md_config = parse_config(md_config, mode="md")

    sim_dir = Path(md_config.sim_dir)
    sim_dir.mkdir(parents=True, exist_ok=True)
    log_file = sim_dir / "md.log"
    setup_logging(log_file, log_level)
    traj_path = sim_dir / md_config.traj_name

    system, sim_fns = md_setup(model_config, md_config)

    dynamics_checks = []
    if md_config.dynamics_checks:
        check_list = [
            DynamicsChecks(check.model_dump()) for check in md_config.dynamics_checks
        ]
        dynamics_checks.extend(check_list)

    constraints = []
    if md_config.constraints:
        constraint_list = [Constraint(c.model_dump()) for c in md_config.constraints]
        constraints.extend(constraint_list)

    n_steps = int(np.ceil(md_config.duration / md_config.ensemble.dt))

    traj_handler = H5TrajHandler(
        system,
        md_config.sampling_rate,
        md_config.buffer_size,
        traj_path,
        md_config.ensemble.dt,
        properties=md_config.properties,
        h5md_options=md_config.h5md_options.model_dump(),
    )
    # TODO implement correct chunking

    run_sim(
        system,
        sim_fns,
        md_config.ensemble,
        n_steps=n_steps,
        n_inner=md_config.n_inner,
        extra_capacity=md_config.extra_capacity,
        load_momenta=md_config.load_momenta,
        traj_handler=traj_handler,
        rng_key=jax.random.PRNGKey(md_config.seed),
        restart=md_config.restart,
        checkpoint_interval=md_config.checkpoint_interval,
        sim_dir=sim_dir,
        dynamics_checks=dynamics_checks,
        constraints=constraints,
    )



from collections import Counter

import numpy as np
import pytest
from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator
from jax import vmap
from vesin import NeighborList

from apax.data.preprocessing import compute_nl
from apax.layers.distances import disp_fn
from apax.utils.convert import atoms_to_inputs, atoms_to_labels
from apax.utils.data import split_atoms, split_idxs
from apax.utils.random import seed_py_np_tf

# TODO RE-ENABLE LATER
# @pytest.mark.parametrize(
#     "num_data, pbc, calc_results, external_labels",
#     (
#         [5, False, ["energy"], None],
#         [5, False, ["energy", "forces"], None],
#         [5, True, ["energy", "forces"], None],
#         [
#             5,
#             True,
#             ["energy", "forces"],
#             [{
#                 "name": "ma_tensors",
#                 "values": np.random.uniform(low=-1.0, high=1.0, size=(5, 3, 3)),
#             }],
#         ],
#     ),
# )
# def test_input_pipeline(example_atoms, calc_results, num_data, external_labels):
#     batch_size = 2
#     r_max = 6.0

#     if external_labels:
#         for label in external_labels:
#             for a, v in zip(example_atoms, label["values"]):
#                 a.calc.results[label["name"]] = v

#     ds = InMemoryDataset(
#         example_atoms,
#         r_max,
#         batch_size,
#         1,
#         buffer_size=1000,
#     )
#     assert ds.steps_per_epoch() == num_data // batch_size

#     ds = ds.shuffle_and_batch()

#     sample_inputs, sample_labels = next(ds)

#     assert "box" in sample_inputs
#     assert len(sample_inputs["box"]) == batch_size
#     assert len(sample_inputs["box"][0]) == 3

#     assert "numbers" in sample_inputs
#     for i in range(batch_size):
#         assert len(sample_inputs["numbers"][i]) == max(sample_inputs["n_atoms"])

#     assert "idx" in sample_inputs
#     assert len(sample_inputs["idx"][0]) == len(sample_inputs["idx"][1])

#     assert "positions" in sample_inputs
#     assert len(sample_inputs["positions"][0][0]) == 3
#     for i in range(batch_size):
#         assert len(sample_inputs["positions"][i]) == max(sample_inputs["n_atoms"])

#     assert "n_atoms" in sample_inputs
#     assert len(sample_inputs["n_atoms"]) == batch_size

#     assert "energy" in sample_labels
#     assert len(sample_labels["energy"]) == batch_size

#     if "forces" in calc_results:
#         assert "forces" in sample_labels
#         assert len(sample_labels["forces"][0][0]) == 3
#         for i in range(batch_size):
#             assert len(sample_labels["forces"][i]) == max(sample_inputs["n_atoms"])

#     if external_labels:
#         assert "ma_tensors" in sample_labels
#         assert len(sample_labels["ma_tensors"]) == batch_size

#     sample_inputs2, _ = next(ds)
#     assert (sample_inputs["positions"][0][0] != sample_inputs2["positions"][0][0]).all()


@pytest.mark.parametrize(
    "num_data, pbc, calc_results",
    (
        [10, False, ["energy", "forces"]],
        [10, True, ["energy", "forces"]],
    ),
)
def test_split_data(example_atoms_list):
    seed_py_np_tf(1)
    train_idxs1, val_idxs1 = split_idxs(example_atoms_list, 4, 2)
    train_idxs2, val_idxs2 = split_idxs(example_atoms_list, 4, 2)
    assert np.all(train_idxs1 != train_idxs2) and np.all(val_idxs1 != val_idxs2)

    train_atoms1, val_atoms1 = split_atoms(example_atoms_list, train_idxs1, val_idxs1)
    train_atoms2, val_atoms2 = split_atoms(example_atoms_list, train_idxs2, val_idxs2)
    assert np.all(train_atoms1[0].get_positions() != train_atoms2[0].get_positions())
    assert np.all(val_atoms1[0].get_positions() != val_atoms2[0].get_positions())

    seed_py_np_tf(1)
    train_idxs2, val_idxs2 = split_idxs(example_atoms_list, 4, 2)
    assert np.all(train_idxs1 == train_idxs2) and np.all(val_idxs1 == val_idxs2)

    train_atoms2, val_atoms2 = split_atoms(example_atoms_list, train_idxs2, val_idxs2)
    assert np.all(train_atoms1[0].get_positions() == train_atoms2[0].get_positions())
    assert np.all(val_atoms1[0].get_positions() == val_atoms2[0].get_positions())


@pytest.mark.parametrize(
    "num_data, pbc, calc_results",
    (
        [5, False, ["energy", "forces"]],
        [5, True, ["energy", "forces"]],
    ),
)
def test_convert_atoms_to_arrays(example_atoms_list, pbc):
    inputs = atoms_to_inputs(example_atoms_list)
    labels = atoms_to_labels(example_atoms_list)

    assert "positions" in inputs
    assert len(inputs["positions"]) == len(example_atoms_list)

    assert "numbers" in inputs
    assert len(inputs["numbers"]) == len(example_atoms_list)

    assert "box" in inputs
    assert len(inputs["box"]) == len(example_atoms_list)
    if not pbc:
        assert np.all(inputs["box"][0] < 1e-6)

    assert "n_atoms" in inputs
    assert len(inputs["n_atoms"]) == len(example_atoms_list)

    assert "energy" in labels
    assert len(labels["energy"]) == len(example_atoms_list)

    assert "forces" in labels
    assert len(labels["forces"]) == len(example_atoms_list)


@pytest.mark.parametrize(
    "pbc, calc_results, cell",
    (
        [
            True,
            ["energy"],
            np.array([[1.8, 0.1, 0.0], [0.0, 2.5, 0.1], [0.1, 0.0, 2.5]]),
        ],
        [
            True,
            ["energy"],
            np.array([[1.8, 0.0, 0.0], [0.0, 2.5, 0.0], [0.0, 0.0, 2.5]]),
        ],
        [
            True,
            ["energy"],
            np.array([[1.5, 0.0, 0.5], [0.0, 2.5, 0.0], [0.0, 0.5, 1.5]]),
        ],
    ),
)
def test_neighbors_and_displacements(pbc, calc_results, cell):
    r_max = 2.0

    numbers = np.array([1, 1])
    positions = np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]])

    additional_data = {}
    additional_data["pbc"] = pbc
    additional_data["cell"] = cell

    result_shapes = {
        "energy": (np.random.rand() - 5.0) * 10_000,
    }

    atoms = Atoms(numbers=numbers, positions=positions, **additional_data)
    if calc_results:
        results = {}
        for key in calc_results:
            results[key] = result_shapes[key]
        atoms.calc = SinglePointCalculator(atoms, **results)

    inputs = atoms_to_inputs([atoms])
    box = np.asarray(inputs["box"][0])
    positions = np.asarray(inputs["positions"][0])
    idx, offsets = compute_nl(positions, box, r_max)

    Ri = positions[idx[0]]
    Rj = positions[idx[1]]
    displacement = vmap(disp_fn, (0, 0, None, None), 0)
    apax_dr_vec = displacement(Rj, Ri, None, box)
    apax_dr_vec += offsets
    apax_dr_vec = np.asarray(apax_dr_vec)
    apax_dist = np.linalg.norm(apax_dr_vec, axis=1)

    calculator = NeighborList(cutoff=r_max, full_list=True)
    vesin_dists = calculator.compute(
        points=atoms.positions, box=box.T, periodic=True, quantities="d"
    )[0]

    assert Counter(vesin_dists) == Counter(apax_dist)
    assert np.all(vesin_dists < r_max)
    assert np.all(apax_dist < r_max)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines utility functions."""

from functools import partial
from typing import Any, Iterable, Optional, Union

import jax.numpy as jnp
import numpy as onp
from jax import jit, tree_util
from jax.lib import xla_bridge

Array = jnp.ndarray
PyTree = Any

i16 = jnp.int16
i32 = jnp.int32
i64 = jnp.int64

f32 = jnp.float32
f64 = jnp.float64


CUSTOM_SIMULATION_TYPE = []


def register_custom_simulation_type(t: Any):
    global CUSTOM_SIMULATION_TYPE
    CUSTOM_SIMULATION_TYPE += [t]


def check_custom_simulation_type(x: Any) -> bool:
    if type(x) in CUSTOM_SIMULATION_TYPE:
        raise ValueError()


def static_cast(*xs):
    """Function to cast a value to the lowest dtype that can express it."""
    # NOTE(schsam): static_cast is so named because it cannot be jit.
    if xla_bridge.get_backend().platform == "tpu":
        return (jnp.array(x, jnp.float32) for x in xs)
    else:
        return (jnp.array(x, dtype=onp.min_scalar_type(x)) for x in xs)


def register_pytree_namedtuple(cls):
    tree_util.register_pytree_node(
        cls, lambda xs: (tuple(xs), None), lambda _, xs: cls(*xs)
    )


def merge_dicts(a, b, ignore_unused_parameters=False):
    if not ignore_unused_parameters:
        return {**a, **b}

    merged = dict(a)
    for key in merged.keys():
        b_val = b.get(key)
        if b_val is not None:
            merged[key] = b_val
    return merged


@partial(jit, static_argnums=(1,))
def safe_mask(mask, fn, operand, placeholder=0):
    masked = jnp.where(mask, operand, 0)
    return jnp.where(mask, fn(masked), placeholder)


def high_precision_sum(
    X: Array, axis: Optional[Union[Iterable[int], int]] = None, keepdims: bool = False
):
    """Sums over axes at 64-bit precision then casts back to original dtype."""
    if jnp.issubdtype(X.dtype, jnp.integer):
        dtyp = jnp.int64
    elif jnp.issubdtype(X.dtype, jnp.complexfloating):
        dtyp = jnp.complex128
    else:
        dtyp = jnp.float64

    return jnp.array(jnp.sum(X, axis=axis, dtype=dtyp, keepdims=keepdims), dtype=X.dtype)


def maybe_downcast(x):
    if isinstance(x, jnp.ndarray) and x.dtype is jnp.dtype("float64"):
        return x
    return jnp.array(x, f32)


def is_array(x: Any) -> bool:
    return isinstance(x, (jnp.ndarray, onp.ndarray))



import pathlib

import pytest
from flax.training import checkpoints

from apax.md import ASECalculator
from tests.conftest import initialize_model, load_and_dump_config

TEST_PATH = pathlib.Path(__file__).parent.resolve()


def test_model_loading(get_tmp_path, get_sample_input):
    pytest.MonkeyPatch().chdir(get_tmp_path)
    model_config_path = TEST_PATH / "config.yaml"

    model_config = load_and_dump_config(model_config_path, get_tmp_path)

    inputs, _ = get_sample_input

    _, params = initialize_model(model_config, inputs)

    ckpt = {"model": {"params": params}, "epoch": 0}
    ckpt_dir1 = pathlib.Path("models/apax_dummy/best")
    ckpt_dir2 = pathlib.Path("../models/apax_dummy/best")
    ckpt_dir3 = pathlib.Path("models/best")

    ckpt_dirs = [ckpt_dir1, ckpt_dir2, ckpt_dir3]
    experiments = ["apax_dummy", "apax_dummy", ""]
    for ckpt_dir, exp in zip(ckpt_dirs, experiments):
        ckpt_dir.mkdir(exist_ok=True, parents=True)
        model_config.data.experiment = exp
        model_config.dump_config(ckpt_dir.parent)

        checkpoints.save_checkpoint(
            ckpt_dir=ckpt_dir.resolve(),
            target=ckpt,
            step=0,
            overwrite=True,
        )
        # attempt to load models
        _ = ASECalculator(ckpt_dir.parent)



import logging
import pathlib
import typing

import ase
import h5py
import yaml
import znh5md
import zntrack

from apax.md.simulate import run_md
from apax.nodes.model import ApaxBase
from apax.nodes.utils import check_duplicate_keys

log = logging.getLogger(__name__)


class ApaxJaxMD(zntrack.Node):
    """Class to run a more performant JaxMD simulation with a apax Model.

    Parameters
    ----------
    data: list[ase.Atoms]
        MD starting structure
    data_id: int, default = -1
        index of the configuration from the data list to use
    model: ApaxModel
        model to use for the simulation
    repeat: None|int|tuple[int, int, int]
        number of repeats
    config: str
        path to the MD simulation parameter file
    """

    data: list[ase.Atoms] = zntrack.deps()
    data_id: int = zntrack.params(-1)

    model: ApaxBase = zntrack.deps()
    repeat: None | int | tuple[int, int, int] = zntrack.params(None)

    config: str = zntrack.params_path(None)

    sim_dir: pathlib.Path = zntrack.outs_path(zntrack.nwd / "md")
    init_struc_dir: pathlib.Path = zntrack.outs_path(zntrack.nwd / "initial_structure.h5")

    @property
    def parameter(self) -> dict:
        with self.state.fs.open(self.config, "r") as f:
            parameter = yaml.safe_load(f)

        custom_parameters = {
            "sim_dir": self.sim_dir.as_posix(),
            "initial_structure": self.init_struc_dir.as_posix(),
        }
        check_duplicate_keys(custom_parameters, parameter, log)
        parameter.update(custom_parameters)

        return parameter

    def _write_initial_structure(self):
        atoms = self.data[self.data_id]
        if self.repeat is not None:
            atoms = atoms.repeat(self.repeat)
        db = znh5md.IO(self.init_struc_dir.as_posix())
        db.extend([atoms])

    def run(self):
        """Primary method to run which executes all steps of the model training"""
        if not self.state.restarted:
            self._write_initial_structure()

        run_md(self.model.parameter, self.parameter, log_level="info")

    @property
    def frames(self) -> typing.List[ase.Atoms]:
        with self.state.fs.open(self.sim_dir / "md.h5", "rb") as f:
            with h5py.File(f) as file:
                return znh5md.IO(file_handle=file)[:]



import logging
from dataclasses import field
from typing import Any, Callable, Optional, Tuple, Union

import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np
from jax import Array

from apax.layers.descriptor.gaussian_moment_descriptor import GaussianMomentDescriptor
from apax.layers.distances import make_distance_fn
from apax.layers.empirical import EmpiricalEnergyTerm
from apax.layers.masking import mask_by_atom
from apax.layers.properties import stress_times_vol
from apax.layers.readout import AtomisticReadout
from apax.layers.scaling import PerElementScaleShift
from apax.utils.jax_md_reduced import partition
from apax.utils.math import fp64_sum
from apax.utils.transform import make_energy_only_model

DisplacementFn = Callable[[Array, Array], Array]
MDModel = Tuple[partition.NeighborFn, Callable, Callable]

log = logging.getLogger(__name__)


class FeatureModel(nn.Module):
    """Model wraps some submodel (e.g. a descriptor) to supply distance computation."""

    representation: nn.Module = GaussianMomentDescriptor()
    readout: nn.Module = AtomisticReadout()
    should_average: bool = False
    init_box: np.array = field(default_factory=lambda: np.array([0.0, 0.0, 0.0]))
    inference_disp_fn: Any = None
    mask_atoms: bool = True

    def setup(self):
        self.compute_distances = make_distance_fn(self.init_box, self.inference_disp_fn)

    def __call__(
        self,
        R: Array,
        Z: Array,
        neighbor: Union[partition.NeighborList, Array],
        box,
        offsets,
        perturbation=None,
    ):
        dr_vec, idx = self.compute_distances(
            R,
            neighbor,
            box,
            offsets,
            perturbation,
        )

        features = self.representation(dr_vec, Z, idx)

        if self.readout:
            features = jax.vmap(self.readout)(features)

        if self.mask_atoms:
            features = mask_by_atom(features, Z)
        if self.should_average:
            features = jnp.mean(features, axis=0)
        return features


class EnergyModel(nn.Module):
    """Model which post processes the output of an atomistic model and
    adds empirical energy terms.
    """

    representation: nn.Module = GaussianMomentDescriptor()
    readout: nn.Module = AtomisticReadout()
    scale_shift: nn.Module = PerElementScaleShift()
    property_heads: list[nn.Module] = field(default_factory=lambda: [])
    corrections: list[EmpiricalEnergyTerm] = field(default_factory=lambda: [])
    init_box: np.array = field(default_factory=lambda: np.array([0.0, 0.0, 0.0]))
    mask_atoms: bool = True
    inference_disp_fn: Any = None

    def setup(self):
        self.compute_distances = make_distance_fn(self.init_box, self.inference_disp_fn)

    def __call__(
        self,
        R: Array,
        Z: Array,
        neighbor: Union[partition.NeighborList, Array],
        box,
        offsets,
        perturbation=None,
    ):
        dr_vec, idx = self.compute_distances(
            R,
            neighbor,
            box,
            offsets,
            perturbation,
        )

        # Model Core
        # shape Natoms
        # shape shallow ens: Natoms x Nensemble
        g = self.representation(dr_vec, Z, idx)
        h = jax.vmap(self.readout)(g)
        E_i = self.scale_shift(h, Z)

        if self.mask_atoms:
            E_i = mask_by_atom(E_i, Z)

        # check for shallow ensemble
        is_shallow_ensemble = E_i.shape[1] > 1
        if is_shallow_ensemble:  # is this necessary or is using sum with axis=0 enough?
            total_energies_ensemble = fp64_sum(E_i, axis=0)
            # shape Nensemble
            energy = total_energies_ensemble
        else:
            # shape ()
            energy = fp64_sum(E_i)

        properties = {}
        for property_head in self.property_heads:
            result = property_head(g, R, dr_vec, Z, idx, box)
            properties.update(result)

        # Corrections
        for correction in self.corrections:
            energy_correction = correction(R, dr_vec, Z, idx, box, properties)
            energy = energy + energy_correction

        return energy, properties


class EnergyDerivativeModel(nn.Module):
    """Transforms an EnergyModel into one that also predicts derivatives the total energy.
    Can calculate forces and stress tensors.
    """

    # Alternatively, should this be a function transformation?
    energy_model: EnergyModel = EnergyModel()
    calc_stress: bool = False

    def __call__(
        self,
        R: Array,
        Z: Array,
        neighbor: Union[partition.NeighborList, Array],
        box,
        offsets,
    ):
        ef_function = jax.value_and_grad(self.energy_model, has_aux=True)
        (energy, properties), neg_forces = ef_function(R, Z, neighbor, box, offsets)
        forces = -neg_forces
        prediction = {"energy": energy, "forces": forces}
        prediction.update(properties)

        if self.calc_stress:
            stress = stress_times_vol(
                make_energy_only_model(self.energy_model),
                R,
                box,
                Z=Z,
                neighbor=neighbor,
                offsets=offsets,
            )
            prediction["stress"] = stress

        return prediction


def make_mean_energy_fn(energy_fn):
    def mean_energy_fn(
        R: Array,
        Z: Array,
        neighbor: Union[partition.NeighborList, Array],
        box,
        offsets,
        perturbation=None,
    ):
        e_ens, _ = energy_fn(R, Z, neighbor, box, offsets, perturbation)
        E_mean = jnp.mean(e_ens)
        return E_mean

    return mean_energy_fn


def make_member_chunk_jac(energy_model, start, end):
    def energy_chunk_fn(R, Z, neighbor, box, offsets):
        Ei = energy_model(R, Z, neighbor, box, offsets)[start:end]
        return Ei

    grad_i_fn = jax.jacrev(energy_chunk_fn)
    return grad_i_fn


class ShallowEnsembleModel(nn.Module):
    """Transforms an EnergyModel into one that also predicts derivatives the total energy.
    Can calculate forces and stress tensors.
    """

    energy_model: EnergyModel = EnergyModel()
    calc_stress: bool = False
    force_variance: bool = True
    chunk_size: Optional[int] = None

    def __call__(
        self,
        R: Array,
        Z: Array,
        neighbor: Union[partition.NeighborList, Array],
        box,
        offsets,
    ):
        energy_ens, properties = self.energy_model(R, Z, neighbor, box, offsets)
        # The two functions below drop the calculation of properties
        mean_energy_fn = make_mean_energy_fn(self.energy_model)
        energy_fn = make_energy_only_model(self.energy_model)

        n_ens = energy_ens.shape[0]
        divisor = 1 / (n_ens - 1)
        energy_mean = jnp.mean(energy_ens)
        energy_variance = divisor * fp64_sum((energy_ens - energy_mean) ** 2)

        prediction = {
            "energy": energy_mean,
            "energy_ensemble": energy_ens,
            "energy_uncertainty": jnp.sqrt(energy_variance),
        }
        prediction.update(properties)

        if self.force_variance:
            if not self.chunk_size:
                forces_ens = -jax.jacrev(energy_fn)(R, Z, neighbor, box, offsets)
            else:
                with jax.ensure_compile_time_eval():
                    if not n_ens % self.chunk_size == 0:
                        m = "the chunksize needs to be a factor of the number of ensemble members"
                        raise ValueError(m)

                forces_ens = []
                start = 0
                for _ in range(n_ens // self.chunk_size):
                    end = start + self.chunk_size
                    jac_i_fn = make_member_chunk_jac(energy_fn, start, end)
                    force_i = -jac_i_fn(R, Z, neighbor, box, offsets)
                    forces_ens.append(force_i)
                    start = end

                n_atoms = R.shape[0]
                forces_ens = jnp.array(forces_ens)
                forces_ens = np.reshape(forces_ens, (n_ens, n_atoms, 3))

            forces_mean = jnp.mean(forces_ens, axis=0)
            forces_variance = divisor * fp64_sum((forces_ens - forces_mean) ** 2, axis=0)

            prediction["forces"] = forces_mean
            prediction["forces_uncertainty"] = jnp.sqrt(forces_variance)

            forces_ens = jnp.transpose(forces_ens, (1, 2, 0))
            prediction["forces_ensemble"] = forces_ens  # n_atoms x 3 x n_members

        else:
            forces_mean = -jax.grad(mean_energy_fn)(R, Z, neighbor, box, offsets)
            prediction["forces"] = forces_mean

        if self.calc_stress:
            stress = stress_times_vol(
                mean_energy_fn, R, box, Z=Z, neighbor=neighbor, offsets=offsets
            )
            prediction["stress"] = stress

        return prediction



import pathlib

import pytest
from flax.training import checkpoints

from apax.bal.api import kernel_selection
from tests.conftest import initialize_model, load_and_dump_config

TEST_PATH = pathlib.Path(__file__).parent.resolve()


@pytest.mark.parametrize(
    "config",
    ["config.yaml", "config_shallow.yaml"],  # ["config_shallow.yaml"]
)
@pytest.mark.parametrize(
    "features",
    [
        {
            "name": "ll_grad",
            "layer_name": "dense_2",
        },
        {
            "name": "ll_force_feat",
            "strategy": "flatten",
        },
        {
            "name": "full_grad_rp",
            "num_rp": 100,
        },
    ],
)
@pytest.mark.parametrize(
    "num_data, pbc, calc_results",
    ([20, False, ["energy", "forces"]],),
)
def test_kernel_selection(
    config, features, example_atoms_list, get_tmp_path, get_sample_input
):
    model_config_path = TEST_PATH / config  # "config.yaml"

    model_config = load_and_dump_config(model_config_path, get_tmp_path)

    inputs, _ = get_sample_input

    _, params = initialize_model(model_config, inputs)

    ckpt = {"model": {"params": params}, "epoch": 0}
    best_dir = model_config.data.best_model_path
    checkpoints.save_checkpoint(
        ckpt_dir=best_dir,
        target=ckpt,
        step=0,
        overwrite=True,
    )

    num_data = len(example_atoms_list)
    n_train = num_data // 2
    train_atoms = example_atoms_list[:n_train]
    pool_atoms = example_atoms_list[n_train:]

    base_fm_options = features
    selection_method = "max_dist"
    bs = 5

    ranking_indices, distances, g_train, g_pool = kernel_selection(
        model_config.data.model_version_path,
        train_atoms,
        pool_atoms,
        base_fm_options,
        selection_method,
        selection_batch_size=bs,
        processing_batch_size=1,
    )
    assert len(ranking_indices) == bs



import os
import subprocess

import ase.io
import zntrack

from apax.nodes.utils import AddData


def test_add_data(tmp_path, get_md22_stachyose):
    os.chdir(tmp_path)
    subprocess.run(["git", "init"], check=True)
    subprocess.run(["dvc", "init"], check=True)
    proj = zntrack.Project()
    with proj:
        data = AddData(file=get_md22_stachyose, stop=50)

    proj.repro()
    data = data.from_rev()
    assert isinstance(data.frames, list)
    assert len(data.frames) == 50
    assert all(isinstance(atoms, ase.Atoms) for atoms in data.frames)
